from bxcommon.utils.stats.stat_event_logic_flags import StatEventLogicFlags
from bxcommon.utils.stats.stat_event_type_settings import StatEventTypeSettings


# pylint: disable=invalid-name
class TransactionStatEventType:
    TX_RECEIVED_FROM_BLOCKCHAIN_NODE = StatEventTypeSettings("TxReceivedFromBlockchainNode",
                                                             event_logic_flags=StatEventLogicFlags.SUMMARY)
    TX_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN = StatEventTypeSettings("TxReceivedFromBlockchainNodeIgnoreSeen")
    TX_SENT_FROM_GATEWAY_TO_PEERS = StatEventTypeSettings("TxSentFromGatewayToPeers",
                                                          event_logic_flags=StatEventLogicFlags.SUMMARY)
    TX_SENT_FROM_GATEWAY_TO_BLOCKCHAIN_NODE = StatEventTypeSettings(
        "TxSentFromGatewayToBlockchainNode", event_logic_flags=StatEventLogicFlags.SUMMARY, priority=True)
    TX_RECEIVED_BY_GATEWAY_FROM_PEER = StatEventTypeSettings("TxReceivedByGatewayFromPeer",
                                                             event_logic_flags=StatEventLogicFlags.SUMMARY)
    TX_RECEIVED_BY_GATEWAY_FROM_PEER_IGNORE_SEEN = StatEventTypeSettings("TxReceivedByGatewayFromPeerIgnoreSeen")
    TX_RECEIVED_BY_RELAY_FROM_PEER = StatEventTypeSettings("TxReceivedByRelayFromPeer",
                                                           event_logic_flags=StatEventLogicFlags.SUMMARY, priority=True)
    TX_RECEIVED_BY_RELAY_FROM_PEER_IGNORE_SEEN = StatEventTypeSettings(
        "TxReceivedByRelayFromPeerIgnoreSeen", priority=True)
    TX_RECEIVED_BY_RELAY_FROM_PEER_FULL_QUOTA_IGNORED = StatEventTypeSettings(
        "TxReceivedByRelayFromPeerFullQuotaIgnored"
    )
    TX_RECEIVED_BY_RELAY_FROM_PEER_DELAYED = StatEventTypeSettings(
        "TxReceivedByRelayFromPeerDelayed"
    )
    TX_RECEIVED_BY_RELAY_FROM_PEER_IGNORE_EXPIRED = StatEventTypeSettings("TxReceivedByRelayFromPeerIgnoreExpired")
    TX_SHORT_ID_ASSIGNED_BY_RELAY = StatEventTypeSettings("TxShortIdAssignedByRelay")
    TX_SHORT_ID_STORED_BY_GATEWAY = StatEventTypeSettings("TxShortIdStoredByGateway")
    TX_SENT_FROM_RELAY_TO_PEERS = StatEventTypeSettings(
        "TxSentFromRelayToPeers",
        event_logic_flags=StatEventLogicFlags.SUMMARY,
        priority=True
    )
    TX_SENT_FROM_RELAY_TO_PEER_SKIPPED_ULTRA_SLOW = StatEventTypeSettings(
        "TxSentFromRelayToPeerSkippedUltraSlow"
    )
    TX_CONTENTS_REQUEST_SENT_FROM_RELAY = StatEventTypeSettings("TxContentsRequestSentFromRelay")
    TX_CONTENTS_REPLY_SENT_BY_RELAY = StatEventTypeSettings("TxContentsReplySentFromRelay")
    TX_CONTENTS_REPLY_RECEIVED_BY_RELAY = StatEventTypeSettings("TxContentsReplyReceivedByRelay")
    TX_UNKNOWN_SHORT_IDS_REQUESTED_BY_GATEWAY_FROM_RELAY = StatEventTypeSettings(
        "TxUnknownShortIdsRequestedByGatewayFromRelay")
    TX_UNKNOWN_SHORT_IDS_REPLY_RECEIVED_BY_GATEWAY_FROM_RELAY = StatEventTypeSettings(
        "TxUnknownShortIdsReplyReceivedByGatewayFromRelay")
    TX_UNKNOWN_SHORT_IDS_REPLY_SENT_BY_RELAY_TO_GATEWAY = StatEventTypeSettings(
        "TxUnknownShortIdsReplySentByRelayToGateway")
    TX_UNKNOWN_TRANSACTION_FOUND_BY_RELAY = StatEventTypeSettings("TxUnknownTransactionFoundByRelay")
    TX_UNKNOWN_TRANSACTION_NOT_FOUND_BY_RELAY = StatEventTypeSettings("TxUnknownTransactionNotFoundByRelay")
    TX_UNKNOWN_TRANSACTION_RECEIVED_BY_GATEWAY_FROM_RELAY = StatEventTypeSettings(
        "TxUnknownTransactionReceivedByGatewayFromRelay"
    )
    TX_REMOVED_FROM_MEMORY = StatEventTypeSettings("TxRemovedFromMemory", event_logic_flags=StatEventLogicFlags.SUMMARY)
    TX_BLOCK_RECOVERY_STATS = StatEventTypeSettings("TxBlockRecoveryStats")
    TX_RECEIVED_FROM_RPC_REQUEST = StatEventTypeSettings("TxReceivedFromRpcRequest",
                                                                event_logic_flags=StatEventLogicFlags.SUMMARY)
    TX_RECEIVED_FROM_RPC_REQUEST_IGNORE_SEEN = StatEventTypeSettings(
        "TxReceivedFromRpcRequestIgnoreSeen"
    )
    TX_COUNTER_BLOCKED = StatEventTypeSettings("TxCounterBlocked")

    TX_CLOUD_API_REQUEST_RECEIVE = StatEventTypeSettings("TxCloudApiRequestReceive")
    TX_CLOUD_API_REQUEST_SENT = StatEventTypeSettings("TxCloudApiRequestSent")
    TX_CLOUD_API_RESPONSE_RECEIVE = StatEventTypeSettings("TxCloudApiResponseReceive")
    TX_CLOUD_API_RESPONSE_SENT = StatEventTypeSettings("TxCloudApiResponseSent")
    TX_RELAY_HTTPS_SERVER_REQUEST_RECEIVE = StatEventTypeSettings("TxRelayHttpsServerRequestReceive")
    TX_RELAY_HTTPS_SERVER_RESPONSE_SENT = StatEventTypeSettings("TxRelayHttpsServerResponseSent")
    TX_GATEWAY_RPC_RESPONSE_SENT = StatEventTypeSettings("TxGatewayRpcResponseSent")
    TX_RELAY_HTTPS_SERVER_REQUEST_IGNORE_SEEN = StatEventTypeSettings(
        "TxRelayHttpsServerRequestIgnoreSeen"
    )
    TX_VALIDATION_FAILED_GAS_PRICE = StatEventTypeSettings("TxValidationFailedGasPrice")
    TX_VALIDATION_FAILED_STRUCTURE = StatEventTypeSettings("TxValidationFailedStructure")
    TX_VALIDATION_FAILED_SIGNATURE = StatEventTypeSettings("TxValidationFailedSignature")
    TX_FROM_BDN_IGNORE_LOW_GAS_PRICE = StatEventTypeSettings("TxFromBdnIgnoreLowGasPrice")

    TX_BLOCKCHAIN_STATUS = StatEventTypeSettings("TxBlockchainStatus")
    TX_REUSE_SENDER_NONCE = StatEventTypeSettings("TxReuseSenderNonce")



from bxcommon.utils.blockchain_utils.eth import rlp_utils

# pylint: disable=invalid-name
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage


def parse_transaction_bytes(tx_bytes: memoryview) -> TransactionsEthProtocolMessage:
    size = len(tx_bytes)

    txs_prefix = rlp_utils.get_length_prefix_list(size)
    size += len(txs_prefix)

    buf = bytearray(size)

    buf[0:len(txs_prefix)] = txs_prefix
    buf[len(txs_prefix):] = tx_bytes
    return TransactionsEthProtocolMessage(buf)



from typing import TYPE_CHECKING

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.utils.blockchain_utils.btc import btc_common_utils
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxutils import logging
from bxgateway import log_messages

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class BtcRelayConnection(AbstractRelayConnection):

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "AbstractGatewayNode"):
        super(BtcRelayConnection, self).__init__(sock, node)

    def msg_tx(self, msg):
        if msg.tx_val() != TxMessage.EMPTY_TX_VAL:
            hash_val = btc_common_utils.get_txid(msg.tx_val())

            if hash_val != msg.tx_hash():
                self.log_error(log_messages.MALFORMED_TX_FROM_RELAY, hash_val, msg.tx_hash())
                return

        super(BtcRelayConnection, self).msg_tx(msg)



from nacl.exceptions import CryptoError

from bxcommon.exceptions import DecryptionError
from bxcommon.utils import crypto, convert
from bxcommon.utils.crypto import symmetric_decrypt, symmetric_encrypt
from bxcommon.utils.expiration_queue import ExpirationQueue
from bxutils import log_messages
from bxutils import logging

logger = logging.get_logger(__name__)


class EncryptionCacheItem:
    def __init__(self, key, ciphertext, payload) -> None:
        self.key = key
        self.ciphertext = ciphertext
        self.payload = payload

    def decrypt(self):
        if self.key is None or self.ciphertext is None:
            raise DecryptionError("Tried decryption without a key and ciphertext. Can't do that.")
        try:
            self.payload = symmetric_decrypt(self.key, self.ciphertext)
            return bytearray(self.payload)
        except (ValueError, CryptoError) as _e:
            # TODO: need to handle decryption errors, e.g. fake key or ciphertext
            raise DecryptionError("Decryption failed. Key does not match ciphertext.")


class EncryptedCache:
    """
    Storage for in-progress received or sent encrypted blocks.
    """

    def __init__(self, expiration_time_s, alarm_queue) -> None:
        self._cache = {}
        self._expiration_queue = ExpirationQueue(expiration_time_s)
        self._expiration_time_s = expiration_time_s
        self._alarm_queue = alarm_queue

    def encrypt_and_add_payload(self, payload):
        """
        Encrypts payload, computing a hash and storing it along with the key for later release.
        If encryption is disabled for dev, store ciphertext identical to hash_key.
        """
        key, ciphertext = symmetric_encrypt(bytes(payload))
        hash_key = crypto.double_sha256(ciphertext)
        self._add(hash_key, key, ciphertext, payload)
        return ciphertext, hash_key

    def add_ciphertext(self, hash_key, ciphertext):
        if hash_key in self._cache:
            self._cache[hash_key].ciphertext = ciphertext
        else:
            self._add(hash_key, None, ciphertext, None)

    def add_key(self, hash_key, encryption_key):
        if hash_key in self._cache:
            self._cache[hash_key].key = encryption_key
        else:
            self._add(hash_key, encryption_key, None, None)

    def decrypt_and_get_payload(self, hash_key, encryption_key):
        """
        Retrieves and decrypts stored ciphertext.
        Returns None if unable to decrypt.
        """
        cache_item = self._cache[hash_key]
        cache_item.key = encryption_key
        return self._safe_decrypt_item(cache_item, hash_key)

    def decrypt_ciphertext(self, hash_key, ciphertext):
        """
        Retrieves and decrypts ciphertext with stored key info. Stores info in cache.
        Returns None if unable to decrypt.
        """

        cache_item = self._cache[hash_key]
        cache_item.ciphertext = ciphertext

        return self._safe_decrypt_item(cache_item, hash_key)

    def get_encryption_key(self, hash_key):
        return self._cache[hash_key].key

    def pop_ciphertext(self, hash_key):
        return self._cache.pop(hash_key).ciphertext

    def has_encryption_key_for_hash(self, hash_key):
        return hash_key in self._cache and self._cache[hash_key].key is not None

    def has_ciphertext_for_hash(self, hash_key):
        return hash_key in self._cache and self._cache[hash_key].ciphertext is not None

    def hash_keys(self):
        return self._cache.keys()

    def encryption_items(self):
        return self._cache.values()

    def remove_item(self, hash_key):
        if hash_key in self._cache:
            del self._cache[hash_key]

    def _add(self, hash_key, encryption_key, ciphertext, payload):
        self._cache[hash_key] = EncryptionCacheItem(encryption_key, ciphertext, payload)
        self._expiration_queue.add(hash_key)
        self._alarm_queue.register_approx_alarm(self._expiration_time_s * 2, self._expiration_time_s,
                                                self._cleanup_old_cache_items)

    def _cleanup_old_cache_items(self):
        self._expiration_queue.remove_expired(remove_callback=self.remove_item)

    def __iter__(self):
        return iter(self._cache)

    def __len__(self):
        return len(self._cache)

    def _safe_decrypt_item(self, cache_item, hash_key):
        try:
            return cache_item.decrypt()
        except DecryptionError:
            failed_ciphertext = self.pop_ciphertext(hash_key)
            logger.warning(log_messages.DECRYPTION_FAILED,
                           convert.bytes_to_hex(hash_key), convert.bytes_to_hex(failed_ciphertext[-4:]))
            return None



import struct

import blxr_rlp as rlp
from bxcommon.exceptions import ParseError
from bxcommon.utils.blockchain_utils.eth import eth_common_constants
from bxcommon.utils.blockchain_utils.eth.crypto_utils import get_padded_len_16
from bxgateway.utils.eth.frame import Frame


def get_frames(msg_type, payload_bytes, protocol_id=eth_common_constants.DEFAULT_FRAME_PROTOCOL_ID,
               window_size=eth_common_constants.DEFAULT_FRAME_SIZE):
    """
    Parses frames from message bytes

    :param msg_type: type of message
    :param payload_bytes: frames payload bytes
    :param protocol_id: protocol id
    :param window_size: frame window size
    :return: list of frames
    """

    payload_mem_view = memoryview(payload_bytes)

    max_frame_payload_size = get_max_frame_payload_size(window_size)

    if len(payload_bytes) <= max_frame_payload_size:
        return [Frame(msg_type, payload_bytes, protocol_id)]

    frames = []
    total_payload_size = eth_common_constants.FRAME_MSG_TYPE_LEN + len(payload_bytes)
    sequence_id = 0

    while (True):
        start_index = sequence_id * max_frame_payload_size
        end_index = (sequence_id + 1) * max_frame_payload_size

        payload = payload_mem_view[start_index:end_index]
        frame = Frame(msg_type if sequence_id == 0 else None,
                      payload,
                      protocol_id,
                      sequence_id=sequence_id,
                      total_payload_size=(total_payload_size if sequence_id == 0 else None),
                      is_chunked=True)

        frames.append(frame)

        if end_index >= len(payload_mem_view):
            break

        sequence_id += 1

    return frames


def get_max_frame_payload_size(window_size):
    """
    Returns max frame payload size for provided window size
    :param window_size: frame window size
    :return: max payload size
    """

    return window_size - eth_common_constants.FRAME_HDR_DATA_LEN - \
           2 * eth_common_constants.FRAME_MAC_LEN - \
           eth_common_constants.FRAME_MSG_TYPE_LEN


def get_full_frame_size(body_size):
    """
    Returns size of full frame for provided frame body size
    :param body_size: frame body size
    :return: size of full frame
    """

    return eth_common_constants.FRAME_HDR_TOTAL_LEN + \
           get_padded_len_16(body_size) + \
           eth_common_constants.FRAME_MAC_LEN


def parse_frame_header(header_bytes):
    """
    Parses frame header
    :param header_bytes: frame header bytes
    :return: tuple (body size, protocol id, sequence id, total payload size)
    """

    sequence_id = None
    total_payload_size = None

    body_size = struct.unpack(">I", b"\x00" + header_bytes[:3])[0]

    try:
        header_data = rlp.decode(header_bytes[3:], sedes=Frame.header_data_serializer, strict=False)
    except rlp.RLPException:
        raise ParseError("Invalid rlp data in frame header")

    protocol_id = header_data[0]

    if len(header_data) > 1:
        sequence_id = header_data[1]

    if len(header_data) == 3:
        total_payload_size = header_data[2]

    return body_size, protocol_id, sequence_id, total_payload_size


def parse_frame_body(body_bytes, has_msg_type):
    """
    Parses frame body

    :param body_bytes: frame body bytes
    :param has_msg_type: message type
    :return: tuple (payload, message type)
    """

    msg_type = None

    if has_msg_type:
        item, end = rlp.codec.consume_item(body_bytes, 0)
        msg_type = rlp.sedes.big_endian_int.deserialize(item)
        payload = body_bytes[end:]
    else:
        payload = body_bytes

    return payload, msg_type



from bxutils.logging.log_level import LogLevel

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage

from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType


class BlockPropagationRequestMessage(AbstractBloxrouteMessage):
    """
    Request for other gateways to encrypt and propagate block message.
    """
    MESSAGE_TYPE = GatewayMessageType.BLOCK_PROPAGATION_REQUEST

    def __init__(self, blob=None, buf=None):
        if buf is None:
            payload_len = len(blob) + constants.CONTROL_FLAGS_LEN
            buf = bytearray(self.HEADER_LENGTH + payload_len)

            off = self.HEADER_LENGTH
            buf[off:off + len(blob)] = blob
        else:
            payload_len = len(buf) - constants.BX_HDR_COMMON_OFF

        self.buf = buf

        super(BlockPropagationRequestMessage, self).__init__(self.MESSAGE_TYPE, payload_len, self.buf)
        self._blob = None

    def log_level(self):
        return LogLevel.DEBUG

    def blob(self):
        if self._blob is None:
            off = self.HEADER_LENGTH
            self._blob = self.buf[off: off + self.payload_len() - constants.CONTROL_FLAGS_LEN]
        return self._blob



import struct
from typing import List, Tuple, Union

from bxcommon import constants
from bxcommon.models.transaction_info import TransactionInfo
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash


def get_serialized_length(transactions_info: List[TransactionInfo]) -> int:
    message_size = constants.UL_INT_SIZE_IN_BYTES + len(transactions_info) * (
        constants.UL_INT_SIZE_IN_BYTES
        + crypto.SHA256_HASH_LEN
        + constants.UL_INT_SIZE_IN_BYTES
    )

    for tx in transactions_info:
        tx_contents = tx.contents
        assert tx_contents is not None
        message_size += len(tx_contents)

    return message_size


def serialize_transactions_info_to_buffer(txs_info: List[TransactionInfo], buffer: bytearray,
                                          offset: int = 0) -> int:
    struct.pack_into("<L", buffer, offset, len(txs_info))
    offset += constants.UL_INT_SIZE_IN_BYTES

    for tx_info in txs_info:
        struct.pack_into("<L", buffer, offset, tx_info.short_id)
        offset += constants.UL_INT_SIZE_IN_BYTES

        tx_hash = tx_info.hash
        assert tx_hash is not None
        assert tx_hash.binary is not None
        buffer[offset:offset + crypto.SHA256_HASH_LEN] = tx_hash.binary
        offset += crypto.SHA256_HASH_LEN

        tx_contents = tx_info.contents
        assert tx_contents is not None
        struct.pack_into("<L", buffer, offset, len(tx_contents))
        offset += constants.UL_INT_SIZE_IN_BYTES

        buffer[offset:offset + len(tx_contents)] = tx_contents
        offset += len(tx_contents)

    return offset


def serialize_transactions_info(transactions_info: List[TransactionInfo]) -> bytearray:
    serialized_bytes = bytearray(get_serialized_length(transactions_info))
    serialize_transactions_info_to_buffer(transactions_info, serialized_bytes, 0)
    return serialized_bytes


def deserialize_transactions_info_from_buffer(
    buffer: Union[bytearray, memoryview],
    offset: int = 0
) -> Tuple[List[TransactionInfo], int]:
    txs = []

    txs_count, = struct.unpack_from("<L", buffer, offset)
    offset += constants.UL_INT_SIZE_IN_BYTES

    for _ in range(txs_count):
        tx_sid, = struct.unpack_from("<L", buffer, offset)
        offset += constants.UL_INT_SIZE_IN_BYTES

        tx_hash = Sha256Hash(buffer[offset:offset + crypto.SHA256_HASH_LEN])
        offset += crypto.SHA256_HASH_LEN

        tx_size, = struct.unpack_from("<L", buffer, offset)
        offset += constants.UL_INT_SIZE_IN_BYTES

        tx = buffer[offset:offset + tx_size]
        offset += tx_size

        txs.append(TransactionInfo(tx_hash, tx, tx_sid))

    return txs, offset


def deserialize_transactions_info(buffer: Union[bytearray, memoryview]) -> List[TransactionInfo]:
    transactions_info, _ = deserialize_transactions_info_from_buffer(buffer, 0)
    return transactions_info



from dataclasses import dataclass
from typing import Optional

from bxgateway.utils.logging.status.connection_state import ConnectionState


@dataclass
class ConnectionInfo:
    ip_address: Optional[str]
    port: Optional[str]
    fileno: Optional[str]
    connection_time: Optional[str]

    def get_connection_state(self) -> ConnectionState:
        return ConnectionState.DISCONNECTED if self.fileno is None else ConnectionState.ESTABLISHED



from enum import auto

from bxcommon.models.serializable_flag import SerializableFlag


class TxBlockchainStatus(SerializableFlag):
    UNKNOWN = auto()
    PROPAGATING = auto()
    REPROPAGATE = auto()
    TX_POOL = auto()
    MULTIPLE_TX_POOL = auto()
    MINED = auto()
    CONFIRMED = auto()
    DROPPED = auto()
    SPEEDUP = auto()
    CANCELED = auto()

    def __str__(self) -> str:
        # pylint: disable=using-constant-test
        if self.name:
            return str(self.name)
        else:
            return super().__str__()



import typing

TRaw = typing.TypeVar("TRaw")
TEncoded = typing.TypeVar("TEncoded")


class ObjectEncoder(typing.Generic[TRaw, TEncoded]):

    def __init__(
            self,
            encode: typing.Callable[[TRaw], TEncoded],
            decode: typing.Callable[[TEncoded], TRaw]
    ):
        self.encode = encode
        self.decode = decode

    @classmethod
    def raw_encoder(cls):
        return cls(lambda x: x, lambda x: x)



from typing import TYPE_CHECKING
import asyncio

from bxcommon import constants
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.exceptions import ParseError
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_blxr_transaction_rpc_request import AbstractBlxrTransactionRpcRequest
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcAccountIdError, RpcBlocked
from bxcommon.utils import convert
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.messages.eth.serializers.transaction import Transaction

from bxutils import logging, log_messages as common_log_messages

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class GatewayBlxrTransactionRpcRequest(AbstractBlxrTransactionRpcRequest["AbstractGatewayNode"]):
    SYNCHRONOUS = rpc_constants.SYNCHRONOUS_PARAMS_KEY
    synchronous: bool = True

    def validate_params(self) -> None:
        params = self.params
        if params is None or not isinstance(params, dict):
            raise RpcInvalidParams(
                self.request_id,
                "Params request field is either missing or not a dictionary type."
            )
        if (
            rpc_constants.TRANSACTION_JSON_PARAMS_KEY in params
            and self.get_network_protocol() == BlockchainProtocol.ETHEREUM
        ):
            tx_json = params[rpc_constants.TRANSACTION_JSON_PARAMS_KEY]
            tx_bytes = Transaction.from_json_with_validation(tx_json).contents().tobytes()
            params[rpc_constants.TRANSACTION_PARAMS_KEY] = tx_bytes.hex()

        super(GatewayBlxrTransactionRpcRequest, self).validate_params()

        if self.SYNCHRONOUS in params:
            synchronous = params[rpc_constants.SYNCHRONOUS_PARAMS_KEY]
            self.synchronous = convert.str_to_bool(str(synchronous).lower(), default=True)
        else:
            self.synchronous = GatewayBlxrTransactionRpcRequest.synchronous

    async def process_request(self) -> JsonRpcResponse:
        params = self.params
        assert isinstance(params, dict)

        account_id = self.get_account_id()
        if not account_id:
            raise RpcAccountIdError(
                self.request_id,
                "Gateway does not have an associated account. Please register the gateway with an account to submit "
                "transactions through RPC."
            )

        transaction_str: str = params[rpc_constants.TRANSACTION_PARAMS_KEY]
        network_num = self.get_network_num()
        if not self.track_flag:
            self.track_flag = TransactionFlag.PAID_TX
        return await self.process_transaction(network_num, account_id, transaction_str)

    async def process_transaction(
        self, network_num: int, account_id: str, transaction_str: str
    ) -> JsonRpcResponse:

        if self.synchronous:
            return await self.post_process_transaction(
                network_num, account_id, self.track_flag, transaction_str
            )
        else:
            asyncio.create_task(
                self.post_process_transaction(
                    network_num, account_id, self.track_flag, transaction_str
                )
            )
        return JsonRpcResponse(
            self.request_id,
            {
                "tx_hash": "not available with async",
            }
        )

    async def post_process_transaction(
        self, network_num: int, account_id: str, transaction_flag: TransactionFlag, transaction_str: str
    ) -> JsonRpcResponse:
        try:
            message_converter = self.node.message_converter
            assert message_converter is not None, "Invalid server state!"
            transaction = message_converter.encode_raw_msg(transaction_str)
            bx_tx = message_converter.bdn_tx_to_bx_tx(transaction, network_num, transaction_flag, account_id)
        except (ValueError, ParseError) as e:
            logger.error(common_log_messages.RPC_COULD_NOT_PARSE_TRANSACTION, e)
            raise RpcInvalidParams(
                self.request_id, f"Invalid transaction param: {transaction_str}"
            )
        tx_service = self.node.get_tx_service()
        tx_hash = bx_tx.tx_hash()
        transaction_key = tx_service.get_transaction_key(tx_hash)
        if (
            tx_service.has_transaction_contents_by_key(transaction_key) or
            tx_service.removed_transaction_by_key(transaction_key)
        ):
            short_id = tx_service.get_short_id_by_key(transaction_key)
            tx_stats.add_tx_by_hash_event(
                tx_hash,
                TransactionStatEventType.TX_RECEIVED_FROM_RPC_REQUEST_IGNORE_SEEN,
                network_num,
                account_id=account_id, short_id=short_id
            )
            tx_json = {
                "tx_hash": str(tx_hash),
            }
            return self.ok(tx_json)
        tx_stats.add_tx_by_hash_event(
            tx_hash,
            TransactionStatEventType.TX_RECEIVED_FROM_RPC_REQUEST,
            network_num,
            account_id=account_id
        )
        if self.node.has_active_blockchain_peer():
            blockchain_tx_message = self.node.message_converter.bx_tx_to_tx(bx_tx)
            self.node.broadcast(
                blockchain_tx_message,
                connection_types=(ConnectionType.BLOCKCHAIN_NODE,)
            )

        # All connections outside of this one is a bloXroute server
        broadcast_peers = self.node.broadcast(
            bx_tx,
            connection_types=(ConnectionType.RELAY_TRANSACTION,)
        )
        tx_stats.add_tx_by_hash_event(
            tx_hash,
            TransactionStatEventType.TX_SENT_FROM_GATEWAY_TO_PEERS,
            network_num,
            peers=broadcast_peers
        )
        tx_stats.add_tx_by_hash_event(
            tx_hash,
            TransactionStatEventType.TX_GATEWAY_RPC_RESPONSE_SENT,
            network_num
        )
        tx_service.set_transaction_contents_by_key(transaction_key, bx_tx.tx_val())
        tx_json = {
            "tx_hash": str(tx_hash),
        }
        if not self.node.account_model.is_account_valid():
            raise RpcAccountIdError(
                self.request_id,
                "The account associated with this gateway has expired. "
                "Please visit https://portal.bloxroute.com to renew your subscription."
            )
        if self.node.quota_level == constants.FULL_QUOTA_PERCENTAGE:
            raise RpcBlocked(
                self.request_id,
                "The account associated with this gateway has exceeded its daily transaction quota."
            )
        else:
            return self.ok(tx_json)

    def get_network_num(self) -> int:
        return self.node.network_num

    def get_account_id(self) -> str:
        return self.node.account_id

    def get_network_protocol(self) -> BlockchainProtocol:
        blockchain_protocol = self.node.opts.blockchain_protocol
        assert blockchain_protocol is not None
        return BlockchainProtocol(blockchain_protocol)



import os
import platform
import socket

PLATFORM_LINUX = "linux"
PLATFORM_MAC = "darwin"
DEFAULT_TEXT_ENCODING = "utf-8"
LISTEN_ON_IP_ADDRESS = "0.0.0.0"
DEFAULT_NODE_BACKLOG: int = 500
LOCALHOST = "127.0.0.1"
MAX_BYTE_VALUE = 255

TAIL_BYTE_COUNT = 2
MAX_TAIL_BYTE_VALUE = (2 ** 8) ** TAIL_BYTE_COUNT - 1

PUBLIC_IP_ADDR_REGEX = r"[0-9]+(?:\.[0-9]+){3}"
PUBLIC_IP_ADDR_RESOLVER = "http://checkip.dyndns.org/"

NODE_CONFIG_FILE = "config.cfg"
BLXR_ENV_VAR = "BLXR_ENV"

HOSTNAME = socket.gethostname()
OS_VERSION = platform.platform()

MANIFEST_PATH = "MANIFEST.MF"
MANIFEST_SOURCE_VERSION = "source_version"
PROTOCOL_VERSION = "protocol_version"
REQUIRED_PARAMS_IN_MANIFEST = [MANIFEST_SOURCE_VERSION]
VERSION_TYPE_LIST = ["dev", "v", "ci"]

# <editor-fold desc="Internal Constants">
ALL_NETWORK_NUM = 0
DEFAULT_NETWORK_NUM = 1

OUTPUT_BUFFER_MIN_SIZE = 65535
OUTPUT_BUFFER_BATCH_MAX_HOLD_TIME = 0.05

FULL_QUOTA_PERCENTAGE = 100

WS_PROVIDER_MAX_QUEUE_SIZE = 5000
WS_MIN_RECONNECT_TIMEOUT_S = 1
WS_RECONNECT_TIMEOUTS = [1, 2, 3, 5, 8, 13]
WS_MAX_CONNECTION_TIMEOUT_S = 2

# The unsigned integer transaction SID representing null.
# If changing, also change in bxapi/constants.py
NULL_TX_SID = 0
NULL_TX_SIDS = {NULL_TX_SID}
NULL_TX_TIMESTAMP = 0
TX_SID_INTERVAL = 10000000
# </editor-fold>

# <editor-fold desc="Connection Management">

# number of tries to resolve network address
NET_ADDR_INIT_CONNECT_TRIES = 3
NET_ADDR_INIT_CONNECT_RETRY_INTERVAL_SECONDS = 2

MAX_CONN_BY_IP = 30

# seconds interval between checking connection stances
CONNECTION_TIMEOUT = 3

MAX_CONNECT_RETRIES = 3
MAX_CONNECT_TIMEOUT_INCREASE = 7

RECV_BUFSIZE = 1024 * 1024
MAX_BAD_MESSAGES = 3
PING_INTERVAL_S = 60
PING_PONG_TRESHOLD = 0.5
# </editor-fold>

# <editor-fold desc="Logging">

MAX_LOGGED_BYTES_LEN = 500 * 1024

# </editor-fold>

# <editor-fold desc="Message Packing Constants">

STARTING_SEQUENCE_BYTES = bytearray(b"\xFF\xFE\xFD\xFC")
STARTING_SEQUENCE_BYTES_LEN = 4
CONTROL_FLAGS_LEN = 1
UL_TINY_SIZE_IN_BYTES = 1
UL_SHORT_SIZE_IN_BYTES = 2
UL_INT_SIZE_IN_BYTES = 4
UL_ULL_SIZE_IN_BYTES = 8
DOUBLE_SIZE_IN_BYTES = 8
UNSIGNED_SHORT_MAX_VALUE = 65535
UNSIGNED_INT_MAX_VALUE = 4294967295
IP_V4_PREFIX = bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff")
IP_V4_PREFIX_LENGTH = 12
IP_ADDR_SIZE_IN_BYTES = 16
MSG_NULL_BYTE = b"\x00"

# bytes of basic message header
BX_HDR_COMMON_OFF = 16

# bytes for storing message type
MSG_TYPE_LEN = 12

QUOTA_FLAG_LEN = UL_TINY_SIZE_IN_BYTES
SID_LEN = UL_INT_SIZE_IN_BYTES
TRANSACTION_FLAG_LEN = UL_SHORT_SIZE_IN_BYTES

SID_EXPIRE_TIME_SECONDS = 3 * 24 * 60 * 60

BLOCK_ENCRYPTED_FLAG_LEN = 1
BROADCAST_TYPE_LEN = 4

NETWORK_NUM_LEN = UL_INT_SIZE_IN_BYTES
VERSION_NUM_LEN = UL_INT_SIZE_IN_BYTES
VERSIONED_HELLO_MSG_MIN_PAYLOAD_LEN = UL_INT_SIZE_IN_BYTES + NETWORK_NUM_LEN + VERSION_NUM_LEN

NODE_ID_SIZE_IN_BYTES = 16
ACCOUNT_ID_SIZE_IN_BYTES = 36

NULL_ENCRYPT_REPEAT_VALUE = "1"  # must be nonzero string character
BLOXROUTE_HELLO_MESSAGES = [b"hello", b"ack"]
HTTP_MESSAGE = b"HTTP"
BITCOIN_MESSAGES = [b"\x03\x00\x00/*\xe0\x00\x00\x00\x00\x00C", b"\xff\x00d\x00\x00\x00\x01"]
# </editor-fold>

# <editor-fold desc="SDN Constants">
SDN_ROOT_URL = "https://127.0.0.1:8080"
SDN_CONTACT_RETRY_SECONDS = 5
MAX_COUNTRY_LENGTH = 30

# Should use extension modules
USE_EXTENSION_MODULES = True

# Should support compact block message
ACCEPT_COMPACT_BLOCK = True

DUMP_MISSING_SHORT_IDS_PATH = "/app/bxrelay/debug/missing-short-ids"


class SdnRoutes:
    nodes = "/nodes"
    node = "/nodes/{0}"
    account = "/account/{0}"
    node_potential_relays_by_network = "/nodes/{0}/{1}/potential-relays"
    node_gateways = "/nodes/{0}/gateways?streaming={1}"
    node_remote_blockchain = "/nodes/{0}/potential-remote-blockchain-peers"
    node_event = "/nodes/{0}/events"
    blockchain_network = "/blockchain-networks/{0}/{1}"
    blockchain_networks = "/blockchain-networks"
    gateway_inbound_connection = "/nodes/{0}/gateway-inbound-connection"
    bdn_services = "/configs/bdn-services"
    quota_status = "/accounts/quota-status"
    gateway_settings = "/nodes/{0}/gateway-settings"


# </editor-fold>

# <editor-fold desc="Stats Recording">
FIRST_STATS_INTERVAL_S = 5 * 60

THROUGHPUT_STATS_INTERVAL_S = 15
THROUGHPUT_STATS_LOOK_BACK = 5

NODE_STATS_INTERVAL_S = 60

INFO_STATS_INTERVAL_S = 60 * 60

# how often the threaded stats services check for termination
THREADED_STATS_SLEEP_INTERVAL_S = 1

THREAD_POOL_WORKER_COUNT = (os.cpu_count() or 1) * 5
THREADED_HTTP_POOL_SLEEP_INTERVAL_S = 60
HTTP_POOL_MANAGER_COUNT = 1

# TODO: turn this number up to 60 minutes after we've done some testing to ensure that this is ok
MEMORY_STATS_INTERVAL_S = 30 * 60
MEMORY_USAGE_INCREASE_FOR_NEXT_REPORT_BYTES = 1024 * 1024 * 1024

# Percentage for transactions that will be logged by stats service. The value should be controlled by SDN in the future.
TRANSACTIONS_BY_HASH_PERCENTAGE_TO_LOG_STATS_FOR = 0.1
TRANSACTIONS_BY_SID_PERCENTAGE_TO_LOG_STATS_FOR = 0.01
ENABLE_TRANSACTIONS_STATS_BY_SHORT_IDS = False
DEFAULT_THREAD_POOL_PARALLELISM_DEGREE = 1
DEFAULT_TX_MEM_POOL_BUCKET_SIZE = 10000

# </editor-fold>

# <editor-fold desc="Timers">
MAX_KQUEUE_EVENTS_COUNT = 1000
CANCEL_ALARMS = 0

# Fast execution timeout on alarm queue
MIN_SLEEP_TIMEOUT = 0.1

REQUEST_EXPIRATION_TIME = 15 * 60  # TODO: Return this value to 1 minute

# Expiration time for encrypted blocks in cache on relays and gateways
BLOCK_CACHE_TIMEOUT_S = 60 * 60

# Duration to warn on if alarm doesn't execute
WARN_ALARM_EXECUTION_DURATION = 0.2
WARN_ALL_ALARMS_EXECUTION_DURATION = 0.5

# Timeout to warn on if alarm executed later than expected
WARN_ALARM_EXECUTION_OFFSET = 5

# Minimal expired transactions clean up task frequency
MIN_CLEAN_UP_EXPIRED_TXS_TASK_INTERVAL_S = 15

# Duration to warn on if message processing takes longer than
WARN_MESSAGE_PROCESSING_S = 0.1

# Expiration time for cache of relayed blocks hashes
RELAYED_BLOCKS_EXPIRE_TIME_S = 6 * 60 * 60

DUMP_REMOVED_SHORT_IDS_INTERVAL_S = 5 * 60
DUMP_REMOVED_SHORT_IDS_PATH = "/app/bxcommon/debug/removed-short-ids"

CLEAN_UP_SEEN_SHORT_IDS_DELAY_S = 10

REMOVED_TRANSACTIONS_HISTORY_EXPIRATION_S = 6 * 60 * 60
REMOVED_TRANSACTIONS_HISTORY_CLEANUP_INTERVAL_S = 10
REMOVED_TRANSACTIONS_HISTORY_LENGTH_LIMIT = 500000

RESPONSIVENESS_CHECK_INTERVAL_S = 1
RESPONSIVENESS_CHECK_DELAY_WARN_THRESHOLD_S = 0.2
MEMORY_STATS_DURATION_WARN_THRESHOLD_S = 0.2
MSG_HANDLERS_CYCLE_DURATION_WARN_THRESHOLD_S = 0.2
MSG_HANDLERS_DURATION_WARN_THRESHOLD_S = 0.5
NETWORK_OPERATION_CYCLE_DURATION_WARN_THRESHOLD_S = 0.2
NETWORK_OPERATION_DURATION_WARN_THRESHOLD_S = 0.5
GC_DURATION_WARN_THRESHOLD = 0.1
TX_SERVICE_SET_TXS_WARN_THRESHOLD = 0.1

SERIALIZED_MESSAGE_CACHE_EXPIRE_TIME_S = 10
USE_SERIALIZED_MESSAGE_CACHE = False

# </editor-fold>

# <editor-fold desc="Default Values">

# Default transactions contents cache maximum size per network number
DEFAULT_TX_CACHE_MEMORY_LIMIT_BYTES = 250 * 1024 * 1024

# Default maximum allowed length of internal message payload
DEFAULT_MAX_PAYLOAD_LEN_BYTES = 1024 * 1024

# cleanup confirmed blocks in this depth
BLOCK_CONFIRMATIONS_COUNT = 4

DEFAULT_BLOCK_HOLD_TIMEOUT = 0.3

TXS_MSG_SIZE = 64000
TXS_SYNC_TASK_DURATION = 0.15
TX_SERVICE_SYNC_TXS_S = 0.01
SENDING_TX_MSGS_TIMEOUT_S = 15 * 60
TX_SERVICE_CHECK_NETWORKS_SYNCED_S = 10 * 60
LAST_MSG_FROM_RELAY_THRESHOLD_S = 30
PING_TIMEOUT_S = 2

GATEWAY_SWAP_RELAYS_LATENCY_THRESHOLD_MS = 3
NODE_LATENCY_THRESHOLD_MS = 2
FASTEST_PING_LATENCY_THRESHOLD_PERCENT = 0.2
UPDATE_TX_SERVICE_FULLY_SYNCED_S = 1
FIRST_TX_SERVICE_SYNC_PROGRESS_S = 1
TX_SERVICE_SYNC_PROGRESS_S = 10
TX_SERVICE_SYNC_RELAY_IN_NETWORKS_S = 30

ALARM_QUEUE_INIT_EVENT = 1

# extensions memory management params
MAX_ALLOCATION_POINTER_COUNT = 10
MAX_COUNT_PER_ALLOCATION = 10

EMPTY_SOURCE_ID = MSG_NULL_BYTE * 16
DECODED_EMPTY_SOURCE_ID = EMPTY_SOURCE_ID.decode()

TRANSACTION_SERVICE_LOG_TRANSACTIONS_INTERVAL_S = 60 * 15
TRANSACTION_SERVICE_TRANSACTIONS_HISTOGRAM_BUCKETS = 36

EMPTY_ACCOUNT_ID = MSG_NULL_BYTE * 36
DECODED_EMPTY_ACCOUNT_ID = EMPTY_ACCOUNT_ID.decode()

# https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#module-urllib3.util.retry
HTTP_REQUEST_RETRIES_COUNT: int = 3
HTTP_REQUEST_BACKOFF_FACTOR: float = 0.5
HTTP_REQUEST_TIMEOUT: int = 5
HTTP_HEADERS = {"Content-Type": "application/json"}

MAX_EVENT_LOOP_TIMEOUT: float = 0.05

MAX_EXPIRED_TXS_TO_REMOVE = 500

# </editor-fold>

# keep constants_local.py file to override settings in the constants file
# this part should be at the bottom of the file
try:
    # pyre-ignore
    from bxcommon.constants_local import *
except ImportError as e:
    pass

DEFAULT_LIST_LOCATION_ORDER = ["NA", "SA", "EU", "OC", "AS", "AF", "AN"]
DEFAULT_NETWORK_NAME = "bxtest"

UNASSIGNED_NETWORK_NUMBER = -1

PING_PONG_REPLY_TIMEOUT_S = 5 * 60

NODE_COUNTRY_ATTRIBUTE_NAME = "country"
NODE_REGION_ATTRIBUTE_NAME = "region"
NODE_COUNTRY_CHINA = "China"

NODE_SHUTDOWN_TIMEOUT_S = 30

# tx gateway sync snapshot interval, 0 to snapshot the whole mempool
GATEWAY_SYNC_TX_THRESHOLD_S = 30 * 60
GATEWAY_SYNC_SYNC_CONTENT = True
GATEWAY_SYNC_BUILD_MESSAGE_THRESHOLD_S = 0.15
GATEWAY_SYNC_MAX_MESSAGE_SIZE_BYTES = 500 * 1024
TX_CONTENT_NO_SID_EXPIRE_S = 60 * 15
TX_SYNC_USE_SNAPSHOT = False

BYTE_TO_MB = 1024 * 1024

MEM_STATS_OBJECT_SIZE_THRESHOLD = 1024 * 1024
MEM_STATS_OBJECT_COUNT_THRESHOLD = 2000

TX_FROM_PUB_API_TO_RELAY_THRESHOLD_S = 2.0
TRANSACTION_STREAMER_ATTRIBUTE_NAME = "tx_streamer"
PRIVATE_IP_ATTRIBUTE_NAME = "using_private_ip"

EPOCH_DATE: str = "1970-01-01"
END_DATE: str = "2999-01-01"

RPC_SUBSCRIBER_MAX_QUEUE_SIZE = 5000

WS_DEFAULT_PORT = 28332

GC_LOW_MEMORY_THRESHOLD = 1024 * 1024 * 1024 * 3
GC_MEDIUM_MEMORY_THRESHOLD = 1024 * 1024 * 1024 * 3.5
GC_HIGH_MEMORY_THRESHOLD = 1024 * 1024 * 1024 * 4

BLOCKS_FAILED_VALIDATION_HISTORY_SIZE = 10

THROTTLE_RECONNECT_TIME_S = 15 * 60
MAX_HIGH_RECONNECT_ATTEMPTS_ALLOWED = 10

WAIT_FOR_SUBSCRIPTION_TIMEOUT = 60

ETHEREUM_BLOCKCHAIN_PROTOCOL = "Ethereum"

ETHEREUM_MAINNET_NETWORK_NAME: str = "Mainnet"
ETHEREUM_BSC_MAINNET_NETWORK_NAME: str = "BSC-Mainnet"
ETHEREUM_POLYGON_MAINNET_NETWORK_NAME: str = "Polygon-Mainnet"
ETHEREUM_RINKEBY_NETWORK_NAME: str = "Rinkeby"
ETHEREUM_MAINNET_NETWORK_NUM: int = 5
ALLOWED_NETWORKS = {"Mainnet": 5}



import socket
import time
from typing import TYPE_CHECKING

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.messages.eth.discovery.eth_discovery_message_factory import eth_discovery_message_factory
from bxgateway.messages.eth.discovery.eth_discovery_message_type import EthDiscoveryMessageType
from bxgateway.messages.eth.discovery.ping_eth_discovery_message import PingEthDiscoveryMessage
from bxutils import logging
from bxgateway import log_messages
from bxcommon.utils.blockchain_utils.eth import eth_common_constants

if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)


class EthNodeDiscoveryConnection(AbstractGatewayBlockchainConnection["EthGatewayNode"]):
    """
    Discovery protocol connection with Ethereum node.
    This connection is used to obtain public key of Ethereum node from Ping message.
    """

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "EthGatewayNode"):
        super(EthNodeDiscoveryConnection, self).__init__(sock, node)

        self.message_handlers = {
            EthDiscoveryMessageType.PING: self.msg_ping,
            EthDiscoveryMessageType.PONG: self.msg_pong
        }

        self.can_send_pings = True
        self.pong_message = None

        self._pong_received = False

        self.send_ping()
        self.node.alarm_queue.register_alarm(eth_common_constants.DISCOVERY_PONG_TIMEOUT_SEC, self._pong_timeout)

        self.hello_messages = [EthDiscoveryMessageType.PING, EthDiscoveryMessageType.PONG]

    def connection_message_factory(self) -> AbstractMessageFactory:
        return eth_discovery_message_factory

    def ping_message(self) -> AbstractMessage:
        return PingEthDiscoveryMessage(
            None,
            self.node.get_private_key(),
            eth_common_constants.P2P_PROTOCOL_VERSION,
            (self.external_ip, self.external_port, self.external_port),
            (socket.gethostbyname(self.peer_ip), self.peer_port, self.peer_port),
            int(time.time()) + eth_common_constants.PING_MSG_TTL_SEC
        )

    def msg_ping(self, msg):
        pass

    def msg_pong(self, msg):
        self.node.set_remote_public_key(self, msg.get_public_key())
        self._pong_received = True

    def _pong_timeout(self):
        if not self._pong_received:
            self.log_warning(log_messages.PONG_MESSAGE_TIMEOUT)
            self.mark_for_close()



from dataclasses import dataclass

from bxcommon.models.config.abstract_config_model import AbstractConfigModel


@dataclass
class StatsConfigModel(AbstractConfigModel):
    # pyre-fixme[8]: Attribute has type `bool`; used as `None`.
    enable_block_stats: bool = None
    # pyre-fixme[8]: Attribute has type `bool`; used as `None`.
    enable_tx_stats: bool = None
    # pyre-fixme[8]: Attribute has type `float`; used as `None`.
    tx_stats_percentage: float = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    number_of_oldest_txs_in_cache_to_log: int = None
    # pyre-fixme[8]: Attribute has type `bool`; used as `None`.
    calculate_actual_obj_size: bool = None



import struct

from bxcommon import constants
from bxcommon.messages.bloxroute.version_message import VersionMessage
from bxcommon.utils import uuid_pack
from bxcommon.utils.stats import message_utils
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType


class GatewayHelloMessage(VersionMessage):
    """
    Hello message type for Gateway-Gateway connections.

    Exchanges ip/port info for registration. Duplicate ip/ports will be dropped.
    Ordering is provided so both nodes drop the same connection (e.g. the one with lower ordering).
    """
    MESSAGE_TYPE = GatewayMessageType.HELLO
    PAYLOAD_LENGTH = (VersionMessage.VERSION_MESSAGE_LENGTH +
                      constants.IP_ADDR_SIZE_IN_BYTES +
                      constants.UL_SHORT_SIZE_IN_BYTES +
                      constants.UL_INT_SIZE_IN_BYTES +
                      constants.NODE_ID_SIZE_IN_BYTES
                      )
    ADDITIONAL_LENGTH = constants.NODE_ID_SIZE_IN_BYTES

    def __init__(self, protocol_version=None, network_num=None, ip=None, port=None, ordering=None, node_id=None, buf=None):
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + self.PAYLOAD_LENGTH)

            off = VersionMessage.BASE_LENGTH

            message_utils.pack_ip_port(buf, off, ip, port)
            off += constants.IP_ADDR_SIZE_IN_BYTES + constants.UL_SHORT_SIZE_IN_BYTES

            struct.pack_into("<L", buf, off, ordering)
            off += constants.UL_INT_SIZE_IN_BYTES
            struct.pack_into("%ss" % constants.NODE_ID_SIZE_IN_BYTES, buf, off, uuid_pack.to_bytes(node_id))
        self.buf = buf
        self._ip = None
        self._port = None
        self._ordering = None
        self._node_id = None
        super(GatewayHelloMessage, self).__init__(self.MESSAGE_TYPE, self.PAYLOAD_LENGTH, protocol_version,
                                                  network_num, buf)

    def _unpack_buffer(self):
        off = VersionMessage.BASE_LENGTH

        self._ip, self._port = message_utils.unpack_ip_port(self._memoryview[off:].tobytes())
        off += constants.IP_ADDR_SIZE_IN_BYTES + constants.UL_SHORT_SIZE_IN_BYTES

        self._ordering, = struct.unpack_from("<L", self._memoryview, off)
        off += constants.UL_INT_SIZE_IN_BYTES
        _node_id, = struct.unpack_from("%ss" % constants.NODE_ID_SIZE_IN_BYTES, self._memoryview, off)
        self._node_id = uuid_pack.from_bytes(_node_id)

    def ip(self):
        if self._ip is None:
            self._unpack_buffer()
        return self._ip

    def port(self):
        if self._port is None:
            self._unpack_buffer()
        return self._port

    def ordering(self):
        if self._ordering is None:
            self._unpack_buffer()
        return self._ordering

    def node_id(self):
        if self._node_id is None:
            self._unpack_buffer()
        return self._node_id

    def __repr__(self):
        return "GatewayHelloMessage<ip: {}, port: {}, ordering: {}>".format(self.ip(), self.port(), self.ordering())



import struct
from typing import Optional, List, Union

from bxcommon import constants
from bxcommon.messages.bloxroute import short_ids_serializer
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash


class GetCompressedBlockTxsMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.GET_COMPRESSED_BLOCK_TXS

    """
    Message used to request information about services with specified short ids.
    Node needs to reply with TxsWithShortIdsMessage
    """

    _network_num: Optional[int] = None
    _block_hash: Optional[Sha256Hash] = None
    _short_ids: Optional[List[int]] = None

    def __init__(
        self,
        network_num: Optional[int] = None,
        block_hash: Optional[Sha256Hash] = None,
        short_ids: Optional[List[int]] = None,
        buf: Optional[Union[bytearray, memoryview]] = None
    ) -> None:

        """
        Constructor. Expects list of short ids or message bytes.

        :param network_num Network number
        :param block_hash Block hash
        :param short_ids: list of short ids
        :param buf: message bytes
        """

        if buf is None:
            buf = self._serialize(network_num, block_hash, short_ids)
            super().__init__(self.MESSAGE_TYPE, len(buf) - self.HEADER_LENGTH, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._payload_len = None
            self._payload = None

        self._network_num = None
        self._block_hash = None
        self._short_ids = None

    def get_short_ids(self) -> List[int]:
        if self._short_ids is None:
            self._parse()

        short_ids = self._short_ids
        assert short_ids is not None
        return short_ids

    def block_hash(self) -> Sha256Hash:
        if self._block_hash is None:
            self._parse()

        block_hash = self._block_hash
        assert block_hash is not None
        return block_hash

    def network_num(self) -> int:
        if self._network_num is None:
            self._parse()

        network_num = self._network_num
        assert network_num is not None
        return network_num

    def get_serialized_short_ids(self):
        return self.rawbytes()[
               self.HEADER_LENGTH + constants.UL_INT_SIZE_IN_BYTES + crypto.SHA256_HASH_LEN:
               -constants.CONTROL_FLAGS_LEN]

    def _serialize(self, network_num, block_hash, short_ids):
        msg_size = (
            constants.STARTING_SEQUENCE_BYTES_LEN
            + constants.BX_HDR_COMMON_OFF
            + constants.UL_INT_SIZE_IN_BYTES
            + crypto.SHA256_HASH_LEN
            + short_ids_serializer.get_serialized_length(len(short_ids))
            + constants.CONTROL_FLAGS_LEN
        )

        buf = bytearray(msg_size)

        off = self.HEADER_LENGTH

        struct.pack_into("<L", buf, off, network_num)
        off += constants.UL_INT_SIZE_IN_BYTES

        buf[off:off + crypto.SHA256_HASH_LEN] = block_hash.binary
        off += crypto.SHA256_HASH_LEN

        short_ids_serializer.serialize_short_ids_to_buffer(short_ids, buf, off)

        return buf

    def _parse(self):
        off = self.HEADER_LENGTH

        self._network_num, = struct.unpack_from("<L", self.buf, off)
        off += constants.UL_INT_SIZE_IN_BYTES

        self._block_hash = Sha256Hash(self.buf[off: off + crypto.SHA256_HASH_LEN])
        off += crypto.SHA256_HASH_LEN

        short_ids, _ = short_ids_serializer.deserialize_short_ids_from_buffer(self.buf, off)

        self._short_ids = short_ids

    def __repr__(self):
        return "GetCompressedBlockTxsMessage<num_short_ids: {}>".format(len(self.get_short_ids()))

    def __iter__(self):
        for short_id in self.get_short_ids():
            yield short_id

    def __len__(self):
        return len(self.get_short_ids())



from dataclasses import dataclass
from typing import Optional

from bxgateway.utils.logging.status.installation_type import InstallationType


@dataclass
class Environment:
    installation_type: Optional[InstallationType] = None
    platform: Optional[str] = None
    python_version: Optional[str] = None
    python_path: Optional[str] = None



from typing import List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    # pylint: disable=R0401
    from bxcommon.connections.abstract_connection import AbstractConnection

SUFFIXES = ["bytes", "kB", "MB", "GB"]


def byte_count(num_bytes: float) -> str:
    i = 0
    while num_bytes > 1024 and i < len(SUFFIXES):
        num_bytes /= 1024
        i += 1

    return f"{int(num_bytes)} {SUFFIXES[i]}"


def connections(conns: List["AbstractConnection"]) -> str:
    """
    Formats list of connections to a string logged in stats
    :param conns: list of connections
    :return: formatted string
    """

    return ", ".join(conn.format_connection_desc for conn in conns if conn is not None)


def connection(conn: Optional["AbstractConnection"]) -> str:
    """
    Formats connection to a string logged in stats
    :param conn: Connection
    :return: formatted string
    """

    if conn is None:
        return "<None>"

    return conn.format_connection_desc


def duration(duration_ms: float) -> str:
    """
    Formats duration into a string logged in stats

    :param duration_ms: Duration in milliseconds
    :return: formatted duration
    """

    return "{:.2f}ms".format(duration_ms)


def timespan(start_timestamp_s: float, end_timestamp_s: float) -> str:
    """
    Calculates and formats timespan between start and end, and formats to a string logged in stats

    :param start_timestamp_s: start timestamp in seconds
    :param end_timestamp_s: end timestamp in seconds
    :return: formatted string
    """

    duration_ms = (end_timestamp_s - start_timestamp_s) * 1000
    return duration(duration_ms)


def percentage(percent: float) -> str:
    """
    Formats percentage into a string logged in stats

    :param percent: Percentage
    :return: formatted duration
    """

    return "{:.2f}%".format(percent)


def ratio(first_value: float, second_value: float) -> str:
    """
    Calculates ratio of two values and formats the result
    :param first_value: first value
    :param second_value: second value
    :return: formatted ratio
    """
    return percentage(100 - float(first_value) / second_value * 100)



import os
import struct
from abc import ABCMeta

from bxcommon.utils import convert
from bxcommon.utils.crypto import SHA256_HASH_LEN

# Used to take the last few characters of the SHA256 encryption as the hash function.
# This is done because using the last characters of the SHA256 function provides major speed boosts.
PARTIAL_HASH_LENGTH = 4


class AbstractObjectHash:
    """
    Base class for representing hash as an object
    binary is a memoryview or a bytearray
    Assumes that binary does not mutate
    """
    __meta__ = ABCMeta

    def __init__(self, binary) -> None:
        self.binary = bytearray(binary) if isinstance(binary, memoryview) else binary

        self._hash = struct.unpack("<L", self.binary[-PARTIAL_HASH_LENGTH:])[0]

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return isinstance(other, AbstractObjectHash) and self.binary == other.binary

    def __lt__(self, other):
        return other is None or self.binary < other.binary

    def __getitem__(self, arg):
        return self.binary.__getitem__(arg)


class Sha256Hash(AbstractObjectHash):
    """
    Represents SHA256 hash as an object
    binary is a memoryview or a bytearray
    Assumes that binary does not mutate
    """

    def __init__(self, binary) -> None:
        if len(binary) != SHA256_HASH_LEN:
            raise ValueError("Binary has the wrong length.")

        super(Sha256Hash, self).__init__(binary)

    def __repr__(self):
        return "Sha256Hash<binary: {}>".format(convert.bytes_to_hex(self.binary))

    def __str__(self):
        return convert.bytes_to_hex(self.binary)

    def to_string(self, prefix: bool = False) -> str:
        if prefix:
            return f"0x{str(self)}"
        else:
            return str(self)

    @classmethod
    def from_string(cls, s: str) -> "Sha256Hash":
        if s.startswith("0x"):
            s = s[2:]

        return cls(convert.hex_to_bytes(s))

    @classmethod
    def generate_object_hash(cls) -> "Sha256Hash":
        raw_bytes = bytearray(0)
        raw_bytes.extend(os.urandom(SHA256_HASH_LEN))
        return Sha256Hash(raw_bytes)


class ConcatHash(AbstractObjectHash):
    """
    Hash value that is concatenated with additional data.
    binary is a memoryview or a bytearray that is not mutable
    hashstart is the start of the random bytes that we take a hash with.
    """

    def __init__(self, binary, hashstart) -> None:
        super(ConcatHash, self).__init__(binary)

        self._hash = struct.unpack("<L", self.binary[hashstart:hashstart + PARTIAL_HASH_LENGTH])[0]

    def __repr__(self):
        return "ConcatHash<binary: {}>".format(convert.bytes_to_hex(self.binary))

    def __str__(self):
        return convert.bytes_to_hex(self.binary)


NULL_SHA256_HASH = Sha256Hash(binary=bytearray(SHA256_HASH_LEN))



import base64
import ipaddress
import sys
import dataclasses

from typing import Dict, List, Set, Optional, Iterable
from dataclasses import dataclass

from datetime import datetime
from argparse import Namespace
from urllib import parse

from bxutils import logging
from bxutils.logging.log_level import LogLevel
from bxutils.logging.log_format import LogFormat
from bxutils.logging import LoggerConfig
from bxcommon.utils import ip_resolver

from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.models.bdn_account_model_base import BdnAccountModelBase
from bxcommon.models.node_type import NodeType

from bxcommon import constants


logger = logging.get_logger(__name__)


@dataclass()
class SourceFeedOpts:
    source_feed_ip: str
    source_feed_port: int
    source_feed_rpc_user: str
    source_feed_rpc_password: str

    def get_source_feed_auth(self) -> str:
        return base64.b64encode(
            f"{self.source_feed_rpc_user}:{self.source_feed_rpc_password}".encode("utf-8")
        ).decode("utf-8")


@dataclass
class CommonOpts:
    external_ip: str
    external_port: int
    non_ssl_port: int
    continent: str
    country: str
    region: str
    hostname: str
    sdn_url: str
    log_level: LogLevel
    log_format: LogFormat
    log_flush_immediately: bool
    log_level_overrides: Dict[str, LogLevel]
    log_fluentd_enable: bool
    log_fluentd_host: str
    node_id: str
    transaction_pool_memory_limit: float
    info_stats_interval: int
    throughput_stats_interval: int
    memory_stats_interval: int
    dump_detailed_report_at_memory_usage: int
    dump_removed_short_ids: bool
    dump_removed_short_ids_path: str
    enable_buffered_send: bool
    use_extensions: bool
    import_extensions: bool
    thread_pool_parallelism_degree: int
    tx_mem_pool_bucket_size: int
    source_version: str
    ca_cert_url: str
    private_ssl_base_url: str
    log_fluentd_queue_size: int
    log_level_fluentd: LogLevel
    log_level_stdout: LogLevel
    sync_tx_service: bool
    data_dir: str
    transaction_validation: bool

    rpc: bool
    rpc_port: int
    rpc_host: str
    rpc_user: str
    rpc_password: str
    rpc_use_ssl: bool
    rpc_ssl_base_url: str

    # set by node runner
    blockchain_networks: Dict[int, BlockchainNetworkModel]
    split_relays: bool
    blockchain_network_num: int
    outbound_peers: Set[OutboundPeerModel]
    node_type: NodeType
    logger_names: Optional[Iterable[str]]
    third_party_loggers: Optional[List[LoggerConfig]]
    using_private_ip_connection: bool

    # set after node runner
    has_fully_updated_tx_service: bool
    receive_buffer_size: int

    # hard coded configuration values
    stats_calculate_actual_size: bool
    log_detailed_block_stats: bool

    # set according to default values
    node_start_time: datetime
    os_version: str
    sid_expire_time: int
    block_compression_debug: bool
    enable_tcp_quickack: bool

    @classmethod
    def opts_defaults(cls, opts: Namespace) -> Namespace:
        opts.blockchain_networks = {}
        opts.split_relays = True
        opts.stats_calculate_actual_size = False
        opts.log_detailed_block_stats = False
        opts.blockchain_network_num = 0
        opts.outbound_peers = set()
        opts.node_type = NodeType.EXTERNAL_GATEWAY
        opts.has_fully_updated_tx_service = False
        opts.logger_names = []
        opts.third_party_loggers = None
        opts.node_start_time = datetime.utcnow()
        opts.os_version = constants.OS_VERSION
        opts.sid_expire_time = constants.SID_EXPIRE_TIME_SECONDS
        opts.enable_tcp_quickack = True
        opts.using_private_ip_connection = False
        opts.receive_buffer_size = constants.RECV_BUFSIZE

        return opts

    @classmethod
    def from_opts(cls, opts: Namespace):
        field_names = [field.name for field in dataclasses.fields(cls)]

        return cls(
            **{k: v for k, v in cls.opts_defaults(opts).__dict__.items() if k in field_names})

    def __post_init__(self) -> None:
        self.validate_external_ip()

    def validate_external_ip(self) -> None:
        parsed_sdn_url = parse.urlparse(self.sdn_url)
        sdn_host = parsed_sdn_url.netloc.split(":")[0]
        try:
            sdn_ip = ipaddress.ip_address(ip_resolver.blocking_resolve_ip(sdn_host))
            if not sdn_ip.is_private and self.external_ip and ipaddress.ip_address(self.external_ip).is_private:
                logger.fatal(
                    "The specified external IP ({}) is a known private IP address. Try omitting this argument",
                    self.external_ip,
                    exc_info=False
                )
                sys.exit(1)
        # pylint: disable=broad-except
        except Exception:
            logger.debug("SDN might be offline, skipping validate external ip")

    def validate_network_opts(self) -> None:
        pass

    def set_account_options(self, account_model: BdnAccountModelBase) -> None:
        pass



from typing import Union, Optional, Dict, Callable

from bxcommon.exceptions import ParseError
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.utils.blockchain_utils.btc import btc_common_utils
from bxcommon.utils.blockchain_utils.eth import eth_common_utils
from bxcommon.utils.blockchain_utils.ont import ont_common_utils
from bxcommon import constants

protocol_to_bdn_tx_to_bx_tx: Dict[BlockchainProtocol, Callable[
    [Union[bytes, bytearray, memoryview], int, Optional[TransactionFlag], str],
    TxMessage
]]
network_num_to_protocol: Dict[int, str]


def init(blockchain_networks: Dict[int, BlockchainNetworkModel]):
    # pylint: disable=global-statement
    global network_num_to_protocol
    network_num_to_protocol = {}
    for network_num, blockchain_network in blockchain_networks.items():
        network_num_to_protocol[network_num] = blockchain_network.protocol.lower()


def bdn_tx_to_bx_tx(
    raw_tx: Union[bytes, bytearray, memoryview],
    network_num: int,
    transaction_flag: Optional[TransactionFlag] = None,
    account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID
) -> TxMessage:
    try:
        current_protocol = network_num_to_protocol[network_num]
    except KeyError:
        raise ValueError(f"{network_num} does not exist.")
    blockchain_protocol = BlockchainProtocol(current_protocol)
    return protocol_to_bdn_tx_to_bx_tx[blockchain_protocol](raw_tx, network_num, transaction_flag, account_id)


def btc_bdn_tx_to_bx_tx(
    raw_tx: Union[bytes, bytearray, memoryview],
    network_num: int,
    transaction_flag: Optional[TransactionFlag] = None,
    account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID
) -> TxMessage:
    if isinstance(raw_tx, bytes):
        raw_tx = bytearray(raw_tx)
    try:
        tx_hash = btc_common_utils.get_txid(raw_tx)
    except IndexError:
        raise ValueError("Invalid raw transaction provided.")
    return TxMessage(
        message_hash=tx_hash,
        network_num=network_num,
        tx_val=raw_tx,
        transaction_flag=transaction_flag,
        account_id=account_id
    )


def eth_bdn_tx_to_bx_tx(
    raw_tx: Union[bytes, bytearray, memoryview],
    network_num: int,
    transaction_flag: Optional[TransactionFlag] = None,
    account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID
) -> TxMessage:
    if isinstance(raw_tx, bytes):
        raw_tx = bytearray(raw_tx)
    bx_tx, _, tx_item_length, tx_item_start = eth_common_utils.raw_tx_to_bx_tx(
        raw_tx, 0, network_num, transaction_flag, account_id
    )
    if tx_item_length + tx_item_start != len(raw_tx):
        raise ParseError(raw_tx)
    return bx_tx


def ont_bdn_tx_to_bx_tx(
    raw_tx: Union[bytes, bytearray, memoryview],
    network_num: int,
    transaction_flag: Optional[TransactionFlag] = None,
    account_id: str = constants.DECODED_EMPTY_ACCOUNT_ID
) -> TxMessage:
    if isinstance(raw_tx, bytes):
        raw_tx = bytearray(raw_tx)
    try:
        tx_hash, _ = ont_common_utils.get_txid(raw_tx)
    except IndexError:
        raise ValueError("Invalid raw transaction provided.")
    return TxMessage(
        message_hash=tx_hash,
        network_num=network_num,
        tx_val=raw_tx,
        transaction_flag=transaction_flag,
        account_id=account_id
    )


protocol_to_bdn_tx_to_bx_tx = {
    BlockchainProtocol.BITCOIN: btc_bdn_tx_to_bx_tx,
    BlockchainProtocol.BITCOINCASH: btc_bdn_tx_to_bx_tx,
    BlockchainProtocol.ETHEREUM: eth_bdn_tx_to_bx_tx,
    BlockchainProtocol.ONTOLOGY: ont_bdn_tx_to_bx_tx

}



from abc import abstractmethod
from typing import Union, Dict, List, Any, TYPE_CHECKING, Optional, TypeVar, Generic

from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.rpc_request_type import RpcRequestType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

Node = TypeVar("Node", bound="AbstractNode")


class AbstractRpcRequest(Generic[Node]):
    method: RpcRequestType
    request_id: Optional[str]
    params: Union[Dict[str, Any], List[Any], None]
    node: Node
    help: Dict[str, Any]

    def __init__(
        self,
        request: BxJsonRpcRequest,
        node: Node,
    ) -> None:
        self.method = request.method
        self.request_id = request.id
        params = request.params
        if params is not None and isinstance(params, dict):
            params = {k.lower(): v for k, v in params.items()}
        self.params = params
        self.node = node
        self.help = {}

        self.validate_params()

    @abstractmethod
    def validate_params(self) -> None:
        pass

    @abstractmethod
    async def process_request(self) -> JsonRpcResponse:
        pass

    def ok(self, result: Optional[Any]) -> JsonRpcResponse:
        return JsonRpcResponse(
            self.request_id,
            result
        )



from bxcommon.messages.bloxroute.bloxroute_message_control_flags import BloxrouteMessageControlFlags
from bxcommon.messages.validation.message_validation_error import MessageValidationError


class ControlFlagValidationError(MessageValidationError):
    is_cancelled_cut_through: bool

    def __init__(self, msg: str, control_flag_byte: int) -> None:
        super(ControlFlagValidationError, self).__init__(msg)
        if BloxrouteMessageControlFlags(control_flag_byte) is BloxrouteMessageControlFlags.NONE:
            self.is_cancelled_cut_through = True
        else:
            self.is_cancelled_cut_through = False



from typing import List, TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.bloxroute.block_holding_message import BlockHoldingMessage
from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.utils import crypto
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import NULL_ONT_BLOCK_HASH, OntObjectHash
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.stat_block_type import StatBlockType
from bxgateway import ont_constants, log_messages
from bxgateway.connections.ont.ont_base_connection_protocol import OntBaseConnectionProtocol
from bxgateway.messages.ont.consensus_ont_message import OntConsensusMessage
from bxgateway.messages.ont.get_data_ont_message import GetDataOntMessage
from bxgateway.messages.ont.get_headers_ont_message import GetHeadersOntMessage
from bxgateway.messages.ont.headers_ont_message import HeadersOntMessage
from bxgateway.messages.ont.inventory_ont_message import InvOntMessage, InventoryOntType
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxgateway.messages.ont.pong_ont_message import PongOntMessage
from bxgateway.messages.ont.ver_ack_ont_message import VerAckOntMessage
from bxgateway.messages.ont.version_ont_message import VersionOntMessage
from bxgateway.utils.errors.message_conversion_error import MessageConversionError

if TYPE_CHECKING:
    from bxgateway.connections.ont.ont_node_connection import OntNodeConnection


class OntNodeConnectionProtocol(OntBaseConnectionProtocol):
    def __init__(self, connection: "OntNodeConnection") -> None:
        super(OntNodeConnectionProtocol, self).__init__(connection)

        connection.message_handlers.update({
            OntMessageType.VERSION: self.msg_version,
            OntMessageType.INVENTORY: self.msg_inv,
            OntMessageType.BLOCK: self.msg_block,
            OntMessageType.TRANSACTIONS: self.msg_tx,
            OntMessageType.GET_BLOCKS: self.msg_proxy_request,
            OntMessageType.GET_HEADERS: self.msg_get_headers,
            OntMessageType.GET_DATA: self.msg_get_data,
            OntMessageType.HEADERS: self.msg_headers,
            OntMessageType.CONSENSUS: self.msg_consensus
        })

        self.ping_interval_s = ont_constants.ONT_PING_INTERVAL_S

        # TODO: re-enable when  block cleanup polling is supported
        # if self.block_cleanup_poll_interval_s > 0:
        #     self.connection.node.alarm_queue.register_alarm(
        #         self.block_cleanup_poll_interval_s,
        #         self._request_blocks_confirmation
        #     )

    def msg_version(self, _msg: VersionOntMessage) -> None:
        self.connection.on_connection_established()
        reply = VerAckOntMessage(self.magic, self.is_consensus)
        self.connection.enqueue_msg(reply)

        self.connection.schedule_pings()

        if self.connection.is_active():
            self.node.on_blockchain_connection_ready(self.connection)

    def msg_inv(self, msg: InvOntMessage) -> None:
        if not self.node.should_process_block_hash():
            return

        contains_block = False
        inventory_requests = []
        block_hashes = []
        inventory_type, item_hashes = msg.inv_type()

        if inventory_type == InventoryOntType.MSG_BLOCK.value:
            for item_hash in item_hashes:
                block_hashes.append(item_hash)
                if item_hash not in self.node.blocks_seen.contents:
                    contains_block = True
                    inventory_requests.append(item_hash)
        else:
            for item_hash in item_hashes:
                inventory_requests.append(item_hash)

        # TODO: mark_blocks_and_request_cleanup

        for req in inventory_requests:
            get_data = GetDataOntMessage(
                magic=msg.magic(),
                inv_type=inventory_type,
                block=req
            )
            self.connection.enqueue_msg(get_data, prepend=contains_block)

        block_queuing_service = self.node.block_queuing_service_manager.get_block_queuing_service(self.connection)
        if block_queuing_service is not None:
            block_queuing_service.mark_blocks_seen_by_blockchain_node(block_hashes)

    def msg_get_data(self, msg: GetDataOntMessage) -> None:
        inventory_type, item_hash = msg.inv_type()
        if inventory_type == InventoryOntType.MSG_BLOCK.value:
            block_stats.add_block_event_by_block_hash(
                item_hash,
                BlockStatEventType.REMOTE_BLOCK_REQUESTED_BY_GATEWAY,
                network_num=self.connection.network_num,
                more_info="Protocol: {}, Network: {}".format(
                    self.node.opts.blockchain_protocol,
                    self.node.opts.blockchain_network
                )
            )
        block_queuing_service = self.node.block_queuing_service_manager.get_block_queuing_service(self.connection)
        if block_queuing_service is not None:
            block_queuing_service.send_block_to_node(item_hash)

    def msg_get_headers(self, msg: GetHeadersOntMessage):
        block_queuing_service = self.node.block_queueing_service_manager.get_block_queuing_service(self.connection)
        send_successful = False
        if block_queuing_service is not None:
            send_successful = block_queuing_service.try_send_header_to_node(msg.hash_stop())
        if not send_successful:
            self.msg_proxy_request(msg, self.connection)

    def msg_pong(self, msg: PongOntMessage):
        if msg.height() > self.node.current_block_height:
            reply = GetHeadersOntMessage(self.magic, 1, NULL_ONT_BLOCK_HASH, self.node.current_block_hash)
            self.connection.enqueue_msg(reply)

    def msg_headers(self, msg: HeadersOntMessage):
        if not self.node.should_process_block_hash():
            return

        header = msg.headers()[-1]
        raw_hash = crypto.double_sha256(header)
        reply = GetDataOntMessage(self.magic, InventoryOntType.MSG_BLOCK.value,
                                  OntObjectHash(buf=raw_hash, length=ont_constants.ONT_HASH_LEN))
        self.connection.enqueue_msg(reply)

    def msg_consensus(self, msg: OntConsensusMessage):
        if not self.node.opts.is_consensus:
            return
        if msg.consensus_data_type() != ont_constants.BLOCK_PROPOSAL_CONSENSUS_MESSAGE_TYPE:
            return

        block_hash = msg.block_hash()
        node = self.connection.node
        if not node.should_process_block_hash(block_hash):
            return

        node.block_cleanup_service.on_new_block_received(block_hash, msg.prev_block_hash())
        block_stats.add_block_event_by_block_hash(block_hash,
                                                  BlockStatEventType.BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE,
                                                  network_num=self.connection.network_num,
                                                  broadcast_type=BroadcastMessageType.CONSENSUS,
                                                  more_info="Protocol: {}, Network: {}".format(
                                                      node.opts.blockchain_protocol,
                                                      node.opts.blockchain_network
                                                  ),
                                                  msg_size=len(msg.rawbytes())
                                                  )

        if block_hash in self.connection.node.blocks_seen.contents:
            self.node.on_block_seen_by_blockchain_node(block_hash, self.connection)
            block_stats.add_block_event_by_block_hash(block_hash,
                                                      BlockStatEventType.BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN,
                                                      network_num=self.connection.network_num,
                                                      broadcast_type=BroadcastMessageType.CONSENSUS)
            self.connection.log_info(
                "Discarding duplicate consensus block {} from local blockchain node.",
                block_hash
            )
            return

        node.track_block_from_node_handling_started(block_hash)

        self.connection.log_info(
            "Processing consensus block {} from local blockchain node.",
            block_hash
        )

        # Broadcast BlockHoldingMessage through relays and gateways
        conns = self.node.broadcast(
            BlockHoldingMessage(block_hash, self.node.network_num),
            broadcasting_conn=self.connection,
            prepend_to_queue=True,
            connection_types=(ConnectionType.RELAY_BLOCK, ConnectionType.GATEWAY)
        )
        if len(conns) > 0:
            block_stats.add_block_event_by_block_hash(block_hash,
                                                      BlockStatEventType.BLOCK_HOLD_SENT_BY_GATEWAY_TO_PEERS,
                                                      network_num=self.node.network_num,
                                                      broadcast_type=BroadcastMessageType.CONSENSUS,
                                                      peers=conns
                                                      )

        try:
            bx_block, block_info = self.node.consensus_message_converter.block_to_bx_block(
                msg,
                self.node.get_tx_service(),
                self.node.opts.enable_block_compression,
                self.node.network.min_tx_age_seconds
            )
        except MessageConversionError as e:
            block_stats.add_block_event_by_block_hash(
                e.msg_hash,
                BlockStatEventType.BLOCK_CONVERSION_FAILED,
                network_num=self.connection.network_num,
                broadcast_type=BroadcastMessageType.CONSENSUS,
                conversion_type=e.conversion_type.value
            )
            self.connection.log_error(log_messages.BLOCK_COMPRESSION_FAIL_ONT_CONSENSUS, e.msg_hash, e)
            return

        if block_info.ignored_short_ids:
            self.connection.log_debug(
                "Ignoring {} new SIDs for {}: {}",
                len(block_info.ignored_short_ids), block_info.block_hash, block_info.ignored_short_ids
            )

        block_stats.add_block_event_by_block_hash(block_hash,
                                                  BlockStatEventType.BLOCK_COMPRESSED,
                                                  start_date_time=block_info.start_datetime,
                                                  end_date_time=block_info.end_datetime,
                                                  network_num=self.connection.network_num,
                                                  broadcast_type=BroadcastMessageType.CONSENSUS,
                                                  prev_block_hash=block_info.prev_block_hash,
                                                  original_size=block_info.original_size,
                                                  txs_count=block_info.txn_count,
                                                  blockchain_network=self.node.opts.blockchain_network,
                                                  blockchain_protocol=self.node.opts.blockchain_protocol,
                                                  matching_block_hash=block_info.compressed_block_hash,
                                                  matching_block_type=StatBlockType.COMPRESSED.value,
                                                  more_info="Consensus compression: {}->{} bytes, {}, {}; "
                                                            "Tx count: {}".format(
                                                      block_info.original_size,
                                                      block_info.compressed_size,
                                                      stats_format.percentage(block_info.compression_rate),
                                                      stats_format.duration(block_info.duration_ms),
                                                      block_info.txn_count
                                                  )
                                                  )

        self.node.block_processing_service._process_and_broadcast_compressed_block(
            bx_block, self.connection, block_info, block_hash
        )

    def _build_get_blocks_message_for_block_confirmation(self, hashes: List[Sha256Hash]) -> AbstractMessage:
        raise NotImplementedError



from typing import TYPE_CHECKING

from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class GatewayStopRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {"params": "", "description": "shutdown request to the gateway"}

    def validate_params(self) -> None:
        pass

    async def process_request(self) -> JsonRpcResponse:
        self.node.should_force_exit = True
        return self.ok({})



import socket
import sys
import typing
from abc import abstractmethod
from asyncio import BaseTransport, Transport, BaseProtocol
from typing import TYPE_CHECKING, Optional

from cryptography.x509 import Certificate

from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.network_direction import NetworkDirection
from bxcommon.network.socket_connection_state import SocketConnectionState, SocketConnectionStates
from bxcommon.utils.stats import hooks
from bxutils import logging
from bxutils.logging import LogRecordType
from bxutils.ssl import ssl_certificate_factory

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)
network_troubleshooting_logger = logging.get_logger(LogRecordType.NetworkTroubleshooting, __name__)
SO_QUICKACK = 12


class AbstractSocketConnectionProtocol(BaseProtocol):
    transport: Optional[Transport]
    file_no: int
    endpoint: IpEndpoint
    direction: NetworkDirection
    can_send: bool
    state: SocketConnectionState
    is_ssl: bool

    _node: "AbstractNode"
    _should_retry: bool
    _receive_buf: bytearray

    # performance critical attributes, have been pulled out of state
    alive: bool
    initialized: bool

    def __init__(
        self,
        node: "AbstractNode",
        endpoint: Optional[IpEndpoint] = None,
        is_ssl: bool = True,
    ):
        self._node = node
        self.transport: Optional[Transport] = None
        self.file_no = -1
        # pyre-fixme[8]: Attribute has type `IpEndpoint`; used as
        #  `Optional[IpEndpoint]`.
        self.endpoint = endpoint
        if self.endpoint is None:
            self.direction = NetworkDirection.INBOUND
        else:
            self.direction = NetworkDirection.OUTBOUND
        self.can_send = False
        self.state = SocketConnectionStates.CONNECTING
        self.is_ssl = is_ssl
        self._should_retry = self.direction == NetworkDirection.OUTBOUND
        self._initial_bytes = None
        self._receive_buf = bytearray(node.opts.receive_buffer_size)

        self.alive = True
        self.initialized = False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <{self.endpoint}, {self.direction.name}>"

    def connection_made(self, transport: BaseTransport) -> None:
        self.transport = typing.cast(Transport, transport)

        sock = transport.get_extra_info("socket")
        self.file_no = sock.fileno()
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        if self._node.opts.enable_tcp_quickack:
            self.enable_tcp_quickack()

        if self.direction == NetworkDirection.INBOUND:
            self.endpoint = IpEndpoint(
                *transport.get_extra_info("peername")
            )
            logger.debug("[{}] - accepted connection.", self)

        self._node.on_connection_added(self)
        self.initialized = True
        self.can_send = True
        self.send()
        logger.debug("[{}] - connection established successfully.", self)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        mark_for_close = self.alive
        self.alive = False
        if not self._should_retry:
            self.state |= SocketConnectionStates.DO_NOT_RETRY
        if exc is not None:
            logger.info(
                "[{}] - lost connection due to an error: {}, closing connection, should_retry: {}.",
                self,
                exc,
                self._should_retry,
            )
        else:
            logger.debug(
                "[{}] - lost connection with peer, should_retry: {}.",
                self,
                self._should_retry,
            )
        self._node.on_connection_closed(self.file_no, mark_for_close)

    def pause_writing(self) -> None:
        self.can_send = False
        logger.debug("[{}] - paused writing.", self)

    def resume_writing(self) -> None:
        self.can_send = True
        self.send()
        logger.debug("[{}] - resumed writing.", self)

    def send(self) -> None:
        # TODO: send should buffer together multiple pending buffers and write\
        #  them together to the Transport.
        total_bytes_sent = 0

        conn = self._node.connection_pool.get_by_fileno(self.file_no)

        if not conn:
            return

        while self.is_sendable():
            data = conn.get_bytes_to_send()
            if not data:
                break

            transport = self.transport
            assert transport is not None, "Connection is broken!"
            # note: transport.write() is non blocking and accepts any length of data
            #       even if data is crossing the buffer high limit (will cause a pause)
            bytes_to_send = len(data)
            # pyre-fixme[16]: `Transport` has no attribute `get_write_buffer_limits`.
            buffer_limits = transport.get_write_buffer_limits()
            logger.trace(
                "[{}] - about to send {} bytes, current buffer used {} with limits {}",
                self,
                bytes_to_send,
                transport.get_write_buffer_size(),
                buffer_limits
            )
            transport.write(data)
            conn.advance_sent_bytes(bytes_to_send)
            total_bytes_sent += bytes_to_send

        if total_bytes_sent:
            logger.trace("[{}] - sent {} bytes", self, total_bytes_sent)

    def send_bytes(self, bytes_to_send: typing.Union[memoryview, bytearray]):
        conn = self._node.connection_pool.get_by_fileno(self.file_no)

        if not conn:
            return

        transport = self.transport
        assert transport is not None, "Connection is broken!"

        # pyre-fixme[16]: `Transport` has no attribute `get_write_buffer_limits`.
        buffer_limits = transport.get_write_buffer_limits()
        logger.trace(
            "[{}] - about to send {} bytes, current buffer used {} with limits {}",
            self,
            bytes_to_send,
            transport.get_write_buffer_size(),
            buffer_limits
        )
        transport.write(bytes_to_send)
        len_bytes_to_sent = len(bytes_to_send)
        hooks.add_throughput_event(
            NetworkDirection.OUTBOUND,
            "outgoing",
            len_bytes_to_sent,
            conn.peer_desc,
            conn.peer_id
        )
        logger.trace("[{}] - sent {} bytes", self, len_bytes_to_sent)

    def pause_reading(self) -> None:
        if self.alive:
            assert self.transport is not None, "Connection is broken!"
            self.state |= SocketConnectionStates.HALT_RECEIVE
            logger.debug("[{}] - paused reading.", self)

    def resume_reading(self) -> None:
        if self.alive:
            assert self.transport is not None, "Connection is broken!"
            # pylint bug
            # pylint: disable=invalid-unary-operand-type
            self.state &= ~SocketConnectionStates.HALT_RECEIVE
            logger.debug("[{}] - resumed reading.", self)

    def mark_for_close(self, should_retry: bool = True) -> None:
        if not self.alive:
            return

        self.alive = False
        self._should_retry = should_retry

        transport = self.transport
        assert transport is not None, "Connection is broken!"
        # ideally this would be `transport.close()`, but buffers don't
        # seem to be flushing for 1+ days
        transport.abort()
        logger.debug(
            "[{}] - marked for close, retrying: {}.", self, should_retry
        )

    def is_receivable(self) -> bool:
        return (
            self.alive
            and SocketConnectionStates.HALT_RECEIVE not in self.state
        )

    def is_sendable(self) -> bool:
        """
        Returns if the socket has room on the buffer and can be sent to.

        This function is very frequently called. Avoid doing any sort of complex
        operations, inline function calls, and avoid flags.
        :return:
        """
        return self.alive and self.initialized and self.can_send

    def get_peer_certificate(self) -> Certificate:
        assert self.transport is not None, "Connection is broken!"
        try:
            return ssl_certificate_factory.get_transport_cert(self.transport)
        except ValueError as e:
            raise TypeError("Socket is not SSL type!") from e

    def get_write_buffer_size(self) -> int:
        transport = self.transport
        assert transport is not None, "Connection is broken!"
        if transport.is_closing():
            return 0
        else:
            return transport.get_write_buffer_size()

    def enable_tcp_quickack(self):
        if "linux" in sys.platform:
            sock = self.transport.get_extra_info("socket")
            if sock is None:
                logger.debug("Socket info is None on connection")
            else:
                sock.setsockopt(socket.SOL_SOCKET, SO_QUICKACK, 1)

    @abstractmethod
    def get_last_read_duration_ms(self) -> float:
        pass

    @abstractmethod
    def get_time_since_read_end_ms(self, end_time: float):
        pass



class GatewayStartArgs:
    GATEWAY_RPC_PORT = "gateway_rpc_port"
    GATEWAY_RPC_HOST = "gateway_rpc_host"
    GATEWAY_RPC_USER = "gateway_rpc_user"
    GATEWAY_RPC_PASSWORD = "gateway_rpc_password"
    DEFAULT_TX_FLAG = "default_tx_flag"
    HTTP_INFO_PORT = "http_info_port"
    CONSENSUS_PORT = "consensus_port"
    RELAY_STATE = "relay_state"
    IS_CONSENSUS = "is_consensus"





from mock import MagicMock

from bxgateway.testing import gateway_helpers
from bxcommon.test_utils.mocks.mock_node_ssl_service import MockNodeSSLService
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.constants import LOCALHOST
from bxcommon.test_utils import helpers
from bxcommon.test_utils.mocks.mock_socket_connection import MockSocketConnection


def make_spy_node(gateway_cls, port, **kwargs):
    opts = gateway_helpers.get_gateway_opts(port, **kwargs)
    if opts.use_extensions:
        helpers.set_extensions_parallelism()
    gateway_node = gateway_cls(opts, MockNodeSSLService(gateway_cls.NODE_TYPE, MagicMock()))
    gateway_node.broadcast = MagicMock(wraps=gateway_node.broadcast)
    gateway_node.requester = MagicMock()
    return gateway_node


def make_spy_connection(connection_cls, fileno, port, node, state=ConnectionState.ESTABLISHED):
    gateway_connection = connection_cls(
        MockSocketConnection(fileno, node, ip_address=LOCALHOST, port=port), node
    )
    if state == ConnectionState.ESTABLISHED:
        gateway_connection.on_connection_established()
    else:
        gateway_connection.state = state
    gateway_connection.enqueue_msg = MagicMock()
    return gateway_connection



from typing import Dict, Optional, List, Any, Set, TYPE_CHECKING, Tuple

from bxutils import logging

from bxcommon.feed.feed import Feed, FeedKey
from bxcommon.feed.subscriber import Subscriber

if TYPE_CHECKING:
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)


class FeedManager:
    feeds: Dict[FeedKey, Feed]
    accounts: Dict[str, Dict[int, int]]
    _node: "AbstractNode"

    def __init__(self, node: "AbstractNode") -> None:
        self.feeds = {}
        self.accounts = {}
        self._node = node

    def __contains__(self, item):
        if isinstance(item, FeedKey):
            return item in self.feeds
        else:
            return FeedKey(item) in self.feeds

    def register_feed(self, feed: Feed) -> None:
        if feed.feed_key in self.feeds:
            raise ValueError(
                f"Cannot register two feeds with the same name: {feed.name}"
            )
        self.feeds[feed.feed_key] = feed

    def subscribe_to_feed(self, feed_key: FeedKey, options: Dict[str, Any]) -> Optional[Subscriber]:
        if feed_key in self.feeds:
            subscriber = self.feeds[feed_key].subscribe(options)
            logger.debug(
                "Creating new subscriber ({}) to {}", subscriber.subscription_id, feed_key
            )
            self._node.reevaluate_transaction_streamer_connection()
            account_id = options.get("account_id", None)
            if account_id is not None:
                if account_id in self.accounts:
                    open_feeds = self.accounts.get(account_id)
                    assert open_feeds is not None
                    if feed_key.network_num in open_feeds:
                        open_feeds[feed_key.network_num] += 1
                    else:
                        open_feeds[feed_key.network_num] = 1
                else:
                    self.accounts[account_id] = {feed_key.network_num: 1}
            return subscriber
        else:
            return None

    def unsubscribe_from_feed(
        self,
        feed_key: FeedKey,
        subscriber_id: str,
        account_id: Optional[str] = None
    ) -> Optional[Subscriber]:
        subscriber = self.feeds[feed_key].unsubscribe(subscriber_id)
        if subscriber is not None:
            logger.debug(
                "Unsubscribing subscriber ({}) from {}",
                subscriber.subscription_id,
                feed_key,
            )

            # Subtract the feed from the account counting
            if account_id is not None:
                if account_id in self.accounts:
                    open_feeds = self.accounts.get(account_id)
                    if open_feeds is not None and feed_key.network_num in open_feeds:
                        if open_feeds[feed_key.network_num] == 1:
                            open_feeds.pop(feed_key.network_num)
                            if len(open_feeds) == 0:
                                self.accounts.pop(account_id)
                        else:
                            open_feeds[feed_key.network_num] -= 1

        self._node.reevaluate_transaction_streamer_connection()
        return subscriber

    def publish_to_feed(self, feed_key: FeedKey, message: Any) -> None:
        if feed_key in self.feeds:
            self.feeds[feed_key].publish(message)

    def get_feed_fields(self, feed_key: FeedKey) -> List[str]:
        return self.feeds[feed_key].FIELDS

    def get_feed(self, feed_key: FeedKey):
        return self.feeds.get(feed_key)

    def get_feed_keys(self, network_num: int = 0) -> List[FeedKey]:
        return [key for key in self.feeds if key.network_num == network_num]

    def any_subscribers(self) -> bool:
        return any(feed.subscriber_count() > 0 for feed in self.feeds.values())

    def get_valid_feed_filters(self, feed_key: FeedKey) -> Set[str]:
        return self.feeds[feed_key].FILTERS

    def validate_feed_filters(self, feed_key: FeedKey, filters: str) -> Tuple[str, List[str]]:
        return self.feeds[feed_key].validate_filters(filters)

    def open_feeds_count(self, account_id: str) -> Optional[Dict[int, int]]:
        return self.accounts.get(account_id, {})



import typing
from abc import abstractmethod
from typing import List

from bxcommon import constants
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway import ont_constants
from bxgateway import gateway_constants
from bxgateway.connections.abstract_blockchain_connection_protocol import AbstractBlockchainConnectionProtocol
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.messages.ont.addr_ont_message import AddrOntMessage
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.messages.ont.get_addr_ont_message import GetAddrOntMessage
from bxgateway.messages.ont.inventory_ont_message import InvOntMessage, InventoryOntType
from bxgateway.messages.ont.ont_message_factory import ont_message_factory
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxgateway.messages.ont.ping_ont_message import PingOntMessage
from bxgateway.messages.ont.pong_ont_message import PongOntMessage
from bxgateway.messages.ont.version_ont_message import VersionOntMessage
from bxutils import logging

logger = logging.get_logger(__name__)


class OntBaseConnectionProtocol(AbstractBlockchainConnectionProtocol):
    def __init__(self, connection: AbstractGatewayBlockchainConnection) -> None:
        super(OntBaseConnectionProtocol, self).__init__(
            connection,
            block_cleanup_poll_interval_s=ont_constants.BLOCK_CLEANUP_NODE_BLOCK_LIST_POLL_INTERVAL_S
                                                        )
        self.node = typing.cast("bxgateway.connections.ont.ont_gateway_node.OntGatewayNode", connection.node)
        self.magic = self.node.opts.blockchain_net_magic
        self.version = self.node.opts.blockchain_version
        self.is_consensus = self.node.opts.is_consensus

        connection.hello_messages = ont_constants.ONT_HELLO_MESSAGES
        connection.header_size = ont_constants.ONT_HDR_COMMON_OFF
        connection.message_handlers = {
            OntMessageType.PING: self.msg_ping,
            OntMessageType.PONG: self.msg_pong,
            OntMessageType.GET_ADDRESS: self.msg_getaddr
        }

        version_msg = VersionOntMessage(self.magic, self.version, self.node.opts.blockchain_port,
                                        self.node.opts.http_info_port, self.node.opts.consensus_port,
                                        ont_constants.STARTUP_CAP, self.node.opts.blockchain_nonce, 0,
                                        self.node.opts.relay, self.node.opts.is_consensus,
                                        f"{gateway_constants.GATEWAY_PEER_NAME} "
                                        f"{self.node.opts.source_version}".encode(constants.DEFAULT_TEXT_ENCODING),
                                        self.node.opts.blockchain_services)
        connection.enqueue_msg(version_msg)

    # pyre-fixme[14]: `msg_block` overrides method defined in
    #  `AbstractBlockchainConnectionProtocol` inconsistently.
    def msg_block(self, msg: BlockOntMessage) -> None:
        block_hash = msg.block_hash()

        if not self.node.should_process_block_hash(block_hash):
            return

        if self.node.block_cleanup_service.is_marked_for_cleanup(block_hash):
            self.connection.log_trace("Marked block for cleanup: {}", block_hash)
            self.node.block_cleanup_service.clean_block_transactions(
                transaction_service=self.node.get_tx_service(),
                block_msg=msg
            )
        else:
            self.process_msg_block(msg)

        # After receiving block message sending INV message for the same block to Ontology node
        # This is needed to update Synced Headers value of the gateway peer on the Ontology node
        # If Synced Headers is not up-to-date than Ontology node does not push compact blocks to the gateway
        inv_msg = InvOntMessage(magic=self.node.opts.blockchain_net_magic,
                                inv_type=InventoryOntType.MSG_BLOCK, blocks=[block_hash])
        self.connection.enqueue_msg(inv_msg)
        self.node.update_current_block_height(msg.height(), block_hash)

    def msg_ping(self, msg: PingOntMessage) -> None:
        reply = PongOntMessage(self.magic, msg.height())
        self.connection.enqueue_msg(reply)

    def msg_pong(self, msg: PongOntMessage) -> None:
        pass

    def msg_getaddr(self, _msg: GetAddrOntMessage) -> None:
        reply = AddrOntMessage(self.magic)
        self.connection.enqueue_msg(reply)

    @abstractmethod
    def _build_get_blocks_message_for_block_confirmation(self, hashes: List[Sha256Hash]) -> AbstractMessage:
        pass




from bxgateway.messages.eth.eth_normal_message_converter import EthNormalMessageConverter


def create_eth_message_converter(opts):
    if opts.use_extensions or opts.import_extensions:
        from bxgateway.messages.eth.eth_extension_message_converter import EthExtensionMessageConverter

    # TODO temp - need to remove opts.enable_eth_extensions
    if opts.use_extensions and opts.enable_eth_extensions:
        return EthExtensionMessageConverter()
    else:
        return EthNormalMessageConverter()



from typing import cast

from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.tx_service_sync_txs_message import TxServiceSyncTxsMessage
from bxcommon.messages.bloxroute.txs_serializer import TxContentShortIds
from bxcommon.messages.bloxroute.v15.tx_service_sync_txs_message_v15 import TxServiceSyncTxsMessageV15, \
    TxContentShortIdsV15
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter
from bxcommon.models.transaction_flag import TransactionFlag


class _TxSyncMessageConverterV15(AbstractMessageConverter):
    _MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.TX_SERVICE_SYNC_TXS: TxServiceSyncTxsMessageV15
    }

    _MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.TX_SERVICE_SYNC_TXS: TxServiceSyncTxsMessage
    }

    def convert_to_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE
        if msg_type not in self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING:
            raise ValueError(f"Tried to convert unexpected new message type to v15: {msg_type}")
        msg = cast(TxServiceSyncTxsMessage, msg)

        txs_content_short_ids = msg.txs_content_short_ids()
        txs_content_short_ids_v15 = [
            TxContentShortIdsV15(
                item.tx_hash, item.tx_content, item.short_ids,
                [short_id_flag.get_quota_type() for short_id_flag in item.short_id_flags]
            ) for item in txs_content_short_ids
        ]
        network_num = msg.network_num()

        return TxServiceSyncTxsMessageV15(network_num, txs_content_short_ids_v15)

    def convert_from_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING:
            raise ValueError(f"Tried to convert unexpected old message type to v15: {msg_type}")
        msg = cast(TxServiceSyncTxsMessageV15, msg)

        txs_content_short_ids_v15 = msg.txs_content_short_ids()
        txs_content_short_ids = [
            TxContentShortIds(
                item.tx_hash,
                item.tx_content,
                item.short_ids,
                [TransactionFlag(short_id_flag.value) for short_id_flag in item.short_id_flags]
                )
            for item in txs_content_short_ids_v15
        ]
        network_num = msg.network_num()

        return TxServiceSyncTxsMessage(network_num, txs_content_short_ids)

    def convert_first_bytes_to_older_version(self, first_msg_bytes: memoryview) -> memoryview:
        raise NotImplementedError

    def convert_first_bytes_from_older_version(self, first_msg_bytes: memoryview) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_to_older_version(self, last_msg_bytes: memoryview) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_from_older_version(self, last_msg_bytes: memoryview) -> memoryview:
        raise NotImplementedError

    def get_message_size_change_to_older_version(self) -> int:
        raise NotImplementedError

    def get_message_size_change_from_older_version(self) -> int:
        raise NotImplementedError


tx_sync_message_converter_v15 = _TxSyncMessageConverterV15()



from bxcommon.utils.stats.stat_event_logic_flags import StatEventLogicFlags
from bxcommon.utils.stats.stat_event_type_settings import StatEventTypeSettings


# pylint: disable=invalid-name
class BlockStatEventType:
    BLOCK_ANNOUNCED_BY_BLOCKCHAIN_NODE = StatEventTypeSettings("BlockAnnouncedByBlockchainNode", priority=True)
    BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE = StatEventTypeSettings("BlockReceivedFromBlockchainNode", priority=True)
    BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN = StatEventTypeSettings(
        "BlockReceivedFromBlockchainNodeIgnoreSeen"
    )
    BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_FAILED_VALIDATION = StatEventTypeSettings(
        "BlockReceivedFromBlockchainNodeFailedValidation"
    )
    COMPACT_BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE = StatEventTypeSettings(
        "CompactBlockReceivedFromBlockchainNode", priority=True
    )
    COMPACT_BLOCK_RECEIVED_FROM_BLOCKCHAIN_NODE_IGNORE_SEEN = StatEventTypeSettings(
        "CompactBlockReceivedFromBlockchainNodeIgnoreSeen"
    )
    COMPACT_BLOCK_COMPRESSED_SUCCESS = StatEventTypeSettings("CompactBlockCompressedSuccess")
    COMPACT_BLOCK_COMPRESSED_FAILED = StatEventTypeSettings("CompactBlockCompressedFailed")
    COMPACT_BLOCK_RECOVERY_SUCCESS = StatEventTypeSettings("CompactBlockRecoverySuccess")
    COMPACT_BLOCK_RECOVERY_FAILED = StatEventTypeSettings("CompactBlockRecoveryFailed")
    COMPACT_BLOCK_REQUEST_FULL = StatEventTypeSettings("CompactBlockRequestFull")
    BLOCK_COMPRESSED = StatEventTypeSettings(
        "BlockCompressed",
        event_logic_flags=StatEventLogicFlags.BLOCK_INFO | StatEventLogicFlags.MATCH,
        priority=True
    )
    BLOCK_ENCRYPTED = StatEventTypeSettings(
        "BlockEncrypted", event_logic_flags=StatEventLogicFlags.MATCH
    )
    BLOCK_HOLD_REQUESTED = StatEventTypeSettings("BlockHoldRequested", detailed_stat_event=True)
    BLOCK_HOLD_HELD_BLOCK = StatEventTypeSettings("BlockHoldHeldBlock")
    BLOCK_HOLD_LIFTED = StatEventTypeSettings("BlockHoldLifted", detailed_stat_event=True)
    BLOCK_HOLD_TIMED_OUT = StatEventTypeSettings("BlockHoldTimeout")
    BLOCK_HOLD_SENT_BY_GATEWAY_TO_PEERS = StatEventTypeSettings(
        "BlockHoldSentByGatewayToPeers", detailed_stat_event=True
    )
    BLOCK_HOLD_RECEIVED_BY_RELAY_FROM_PEER = StatEventTypeSettings(
        "BlockHoldReceivedByRelayFromPeer", detailed_stat_event=True
    )
    BLOCK_HOLD_SENT_BY_RELAY_TO_PEERS = StatEventTypeSettings(
        "BlockHoldSentByRelayToPeers", detailed_stat_event=True
    )
    ENC_BLOCK_SENT_FROM_GATEWAY_TO_NETWORK = StatEventTypeSettings(
        "EncBlockSentFromGatewayToNetwork", event_logic_flags=StatEventLogicFlags.SUMMARY
    )
    ENC_BLOCK_CUT_THROUGH_SEND_START = StatEventTypeSettings("EncBlockCutThroughSendStart", priority=True)
    ENC_BLOCK_CUT_THROUGH_SEND_END = StatEventTypeSettings("EncBlockCutThroughSendEnd", priority=True)
    ENC_BLOCK_CUT_THROUGH_RECEIVE_START = StatEventTypeSettings(
        "EncBlockCutThroughReceiveStart",
        event_logic_flags=StatEventLogicFlags.SUMMARY | StatEventLogicFlags.PROPAGATION_START,
        priority=True
    )
    ENC_BLOCK_CUT_THROUGH_RECEIVE_END = StatEventTypeSettings(
        "EncBlockCutThroughReceiveEnd", event_logic_flags=StatEventLogicFlags.PROPAGATION_END,
        priority=True
    )
    ENC_BLOCK_CUT_THROUGH_REDUNDANT_SEND = StatEventTypeSettings("EncBlockCutThroughReceiveEnd", priority=True)
    ENC_BLOCK_CUT_THROUGH_IGNORE_SEEN_BLOCK = StatEventTypeSettings(
        "EncBlockCutThroughIgnoreSeenBlock"
    )
    ENC_BLOCK_CUT_THROUGH_SOURCE_CANCELLED = StatEventTypeSettings("EncBlockCutThroughCancelled")
    ENC_BLOCK_CUT_THROUGH_MANGER_CANCELLED = StatEventTypeSettings(
        "EncBlockCutThroughManagerCancelled"
    )
    ENC_BLOCK_RECEIVED_BY_GATEWAY_FROM_NETWORK = StatEventTypeSettings(
        "EncBlockReceivedByGatewayFromNetwork"
    )
    ENC_BLOCK_GATEWAY_IGNORE_NO_BLOCKCHAIN = StatEventTypeSettings(
        "EncBlockGatewayIgnoreNoBlockchain"
    )
    ENC_BLOCK_DECRYPTED_SUCCESS = StatEventTypeSettings("EncBlockDecryptedSuccess")
    ENC_BLOCK_DECRYPTION_ERROR = StatEventTypeSettings("EncBlockDecryptionError")
    ENC_BLOCK_SENT_BLOCK_RECEIPT = StatEventTypeSettings(
        "EncBlockSentBlockReceipt", detailed_stat_event=True
    )
    ENC_BLOCK_RECEIVED_BLOCK_RECEIPT = StatEventTypeSettings("EncBlockReceivedBlockReceipt")
    ENC_BLOCK_PROPAGATION_NEEDED = StatEventTypeSettings("EncBlockPropagationNeeded")
    ENC_BLOCK_COMPRESSED_TXS_REQUEST_SENT_TO_TX_RELAY = StatEventTypeSettings(
        "EncBlockCompressedTxsRequestSentToTxRelay")
    ENC_BLOCK_COMPRESSED_TXS_REQUEST_PROCESSED_BY_TX_RELAY = StatEventTypeSettings(
        "EncBlockCompressedTxsRequestProcessedByTxRelay")
    ENC_BLOCK_COMPRESSED_TXS_REQUEST_TXS_NOT_FOUND_BY_TX_RELAY = StatEventTypeSettings(
        "EncBlockCompressedTxsRequestTxsNotFoundByTxRelay")
    ENC_BLOCK_COMPRESSED_TXS_SENT_TO_RELAYS = StatEventTypeSettings("EncBlockCompressedTxsSentToRelays")
    ENC_BLOCK_COMPRESSED_TXS_RECEIVED_BY_RELAY = StatEventTypeSettings("EncBlockCompressedTxsReceivedByRelay")
    ENC_BLOCK_COMPRESSED_TXS_SENT_TO_GATEWAYS = StatEventTypeSettings("EncBlockCompressedTxsSentToGateways")
    ENC_BLOCK_COMPRESSED_TXS_RECEIVED_BY_GATEWAY = StatEventTypeSettings("EncBlockCompressedTxsReceivedByGateway")

    BX_BLOCK_PROPAGATION_REQUESTED_BY_PEER = StatEventTypeSettings(
        "BxBlockPropagationRequestedByPeer"
    )
    BLOCK_CUT_THROUGH_SUMMARY = StatEventTypeSettings(
        "BlockCutThroughSummary", event_logic_flags=StatEventLogicFlags.SUMMARY
    )

    BLOCK_TO_ENC_BLOCK_MATCH = StatEventTypeSettings(
        "BlockHashToEncBlockHash", event_logic_flags=StatEventLogicFlags.MATCH
    )
    BLOCK_DECOMPRESSED_IGNORE_SEEN = StatEventTypeSettings(
        "BlockDecompressedIgnoreSeen",
        event_logic_flags=StatEventLogicFlags.BLOCK_INFO | StatEventLogicFlags.MATCH,
        priority=True
    )
    BLOCK_DECOMPRESSED_SUCCESS = StatEventTypeSettings(
        "BlockDecompressedSuccess",
        event_logic_flags=StatEventLogicFlags.BLOCK_INFO | StatEventLogicFlags.MATCH,
        priority=True
    )
    BLOCK_DECOMPRESSED_WITH_UNKNOWN_TXS = StatEventTypeSettings(
        "BlockDecompressedWithUnknownTxs",
        event_logic_flags=StatEventLogicFlags.BLOCK_INFO | StatEventLogicFlags.MATCH,
    )
    BLOCK_DECOMPRESSED_FAILED_VALIDATION = StatEventTypeSettings(
        "BlockDecompressedFailedValidation"
    )
    BLOCK_CONVERSION_FAILED = StatEventTypeSettings("BlockConversionFailed")
    BLOCK_RECOVERY_STARTED = StatEventTypeSettings("BlockRecoveryStarted")
    BLOCK_RECOVERY_REPEATED = StatEventTypeSettings("BlockRecoveryRepeated")
    BLOCK_RECOVERY_COMPLETED = StatEventTypeSettings("BlockRecoveryCompleted")
    BLOCK_RECOVERY_CANCELED = StatEventTypeSettings("BlockRecoveryCanceled")
    BLOCK_SENT_TO_BLOCKCHAIN_NODE = StatEventTypeSettings(
        "BlockSentToBlockchainNode",
        event_logic_flags=StatEventLogicFlags.SUMMARY | StatEventLogicFlags.PROPAGATION_END, priority=True
    )
    BLOCK_HEADER_SENT_TO_BLOCKCHAIN_NODE = StatEventTypeSettings(
        "BlockHeaderSentToBlockchainNode", event_logic_flags=StatEventLogicFlags.SUMMARY
    )
    BLOCK_BODY_SENT_TO_BLOCKCHAIN_NODE = StatEventTypeSettings(
        "BlockBodySentToBlockchainNode", event_logic_flags=StatEventLogicFlags.SUMMARY
    )
    BLOCK_IGNORE_SEEN_BY_BLOCKCHAIN_NODE = StatEventTypeSettings("BlockIgnoreSeenByBlockchainNode")
    BLOCK_IGNORE_DUPLICATE_HEIGHT = StatEventTypeSettings("BlockIgnoreDuplicateHeight")
    ENC_BLOCK_KEY_SENT_FROM_GATEWAY_TO_NETWORK = StatEventTypeSettings(
        "EncBlockKeySentFromGatewayToNetwork", event_logic_flags=StatEventLogicFlags.SUMMARY
    )
    ENC_BLOCK_KEY_RECEIVED_BY_GATEWAY_FROM_NETWORK = StatEventTypeSettings(
        "EncBlockKeyReceivedByGatewayFromNetwork"
    )
    ENC_BLOCK_KEY_SENT_BY_GATEWAY_TO_PEERS = StatEventTypeSettings(
        "EncBlockKeySentByGatewayToPeers", detailed_stat_event=True
    )
    ENC_BLOCK_KEY_RELAY_BROADCAST = StatEventTypeSettings(
        "EncBlockKeyBroadcast", event_logic_flags=StatEventLogicFlags.SUMMARY
    )
    REMOTE_BLOCK_RECEIVED_BY_GATEWAY = StatEventTypeSettings("RemoteBlockReceivedByGateway")
    REMOTE_BLOCK_REQUESTED_BY_GATEWAY = StatEventTypeSettings("RemoteBlockRequestedByGateway")
    BLOCK_COUNTER_BLOCKED = StatEventTypeSettings("BlockCounterBlocked")



from bxcommon.test_utils.abstract_test_case import AbstractTestCase
from bxcommon.test_utils import helpers

from bxcommon.utils.blockchain_utils.eth import crypto_utils, eth_common_constants
from bxgateway.utils.eth.rlpx_cipher import RLPxCipher


class AbstractRLPxCipherTest(AbstractTestCase):
    """
    Base class for unit test that need to use RLPx Cipher
    """

    def setup_ciphers(self, private_key1=None, private_key2=None):
        if private_key1 is None:
            private_key1 = crypto_utils.make_private_key(helpers.generate_bytearray(111))
        if private_key2 is None:
            private_key2 = crypto_utils.make_private_key(helpers.generate_bytearray(111))

        public_key1 = crypto_utils.private_to_public_key(private_key1)
        public_key2 = crypto_utils.private_to_public_key(private_key2)

        cipher1 = RLPxCipher(True, private_key1, public_key2)
        cipher2 = RLPxCipher(False, private_key2, public_key1)

        self.assertFalse(cipher1.is_ready())
        self.assertFalse(cipher2.is_ready())

        auth_msg = cipher1.create_auth_message()
        self.assertTrue(auth_msg)

        enc_auth_msg = cipher1.encrypt_auth_message(auth_msg)

        decrypted_auth_msg, size = cipher2.decrypt_auth_message(enc_auth_msg)
        self.assertEqual(
            size,
            len(decrypted_auth_msg) +
            eth_common_constants.EIP8_AUTH_PREFIX_LEN + eth_common_constants.ECIES_ENCRYPT_OVERHEAD_LENGTH
        )

        cipher2.parse_auth_message(decrypted_auth_msg)

        auth_ack_message = cipher2.create_auth_ack_message()
        enc_auth_ack_message = cipher2.encrypt_auth_ack_message(auth_ack_message)

        cipher1.decrypt_auth_ack_message(enc_auth_ack_message)

        cipher1.setup_cipher()
        cipher2.setup_cipher()

        self.assertTrue(cipher1.is_ready())
        self.assertTrue(cipher2.is_ready())

        return cipher1, cipher2



import struct
import time
import hashlib
from datetime import datetime

from csiphash import siphash24
from collections import deque
from typing import Tuple, Optional, List, Deque, Union, NamedTuple

from bxutils import logging

from bxcommon import constants
from bxcommon.messages.bloxroute import compact_block_short_ids_serializer
from bxcommon.utils import crypto, convert
from bxcommon.messages.bloxroute.compact_block_short_ids_serializer import BlockOffsets
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.blockchain_utils.btc import btc_common_utils
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxcommon.utils.object_hash import Sha256Hash

from bxgateway import btc_constants
from bxgateway.utils.errors import message_conversion_error
from bxgateway.abstract_message_converter import finalize_block_bytes
from bxgateway.messages.btc.abstract_btc_message_converter import AbstractBtcMessageConverter, get_block_info, \
    CompactBlockCompressionResult
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.compact_block_btc_message import CompactBlockBtcMessage
from bxgateway.utils.block_info import BlockInfo
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.utils.block_header_info import BlockHeaderInfo
from bxgateway.messages.btc import btc_messages_util
from bxgateway.abstract_message_converter import BlockDecompressionResult


logger = logging.get_logger(__name__)


class CompactBlockRecoveryData(NamedTuple):
    block_transactions: List[Optional[Union[memoryview, int]]]
    block_header: memoryview
    magic: int
    tx_service: TransactionService


def parse_bx_block_header(
        bx_block: memoryview, block_pieces: Deque[Union[bytearray, memoryview]]
) -> BlockHeaderInfo:
    block_offsets = compact_block_short_ids_serializer.get_bx_block_offsets(bx_block)
    short_ids, short_ids_len = compact_block_short_ids_serializer.deserialize_short_ids_from_buffer(
        bx_block,
        block_offsets.short_id_offset
    )

    # Compute block header hash
    block_header_size = \
        block_offsets.block_begin_offset + \
        btc_constants.BTC_HDR_COMMON_OFF + \
        btc_constants.BTC_BLOCK_HDR_SIZE
    block_hash = BtcObjectHash(
        buf=crypto.bitcoin_hash(
            bx_block[
                block_offsets.block_begin_offset + btc_constants.BTC_HDR_COMMON_OFF:
                block_header_size
            ]
        ),
        length=btc_constants.BTC_SHA_HASH_LEN
    )
    offset = block_header_size

    # Add header piece
    txn_count, txn_count_size = btc_common_utils.btc_varint_to_int(bx_block, block_header_size)
    offset += txn_count_size
    block_pieces.append(bx_block[block_offsets.block_begin_offset:offset])
    return BlockHeaderInfo(block_offsets, short_ids, short_ids_len, block_hash, offset, txn_count)


def parse_bx_block_transactions(
        block_hash: Sha256Hash,
        bx_block: memoryview,
        offset: int,
        short_ids: List[int],
        block_offsets: BlockOffsets,
        tx_service: TransactionService,
        block_pieces: Deque[Union[bytearray, memoryview]]
) -> Tuple[List[int], List[Sha256Hash], int]:
    has_missing, unknown_tx_sids, unknown_tx_hashes = \
        tx_service.get_missing_transactions(short_ids)
    if has_missing:
        return unknown_tx_sids, unknown_tx_hashes, offset
    short_tx_index = 0
    output_offset = offset
    while offset < block_offsets.short_id_offset:
        if bx_block[offset] == btc_constants.BTC_SHORT_ID_INDICATOR:
            try:
                sid = short_ids[short_tx_index]
            except IndexError:
                raise message_conversion_error.btc_block_decompression_error(
                    block_hash,
                    f"Message is improperly formatted, short id index ({short_tx_index}) "
                    f"exceeded its array bounds (size: {len(short_ids)})"
                )
            tx_hash, tx, _ = tx_service.get_transaction(sid)
            offset += btc_constants.BTC_SHORT_ID_INDICATOR_LENGTH
            short_tx_index += 1
        else:
            tx_size = btc_messages_util.get_next_tx_size(bx_block, offset)
            tx = bx_block[offset:offset + tx_size]
            offset += tx_size

        # pyre-fixme[6]: Expected `Union[bytearray, memoryview]` for 1st param but
        #  got `Optional[Union[bytearray, memoryview]]`.
        block_pieces.append(tx)
        # pyre-fixme[6]: Expected `Sized` for 1st param but got
        #  `Optional[Union[bytearray, memoryview]]`.
        output_offset += len(tx)

    return unknown_tx_sids, unknown_tx_hashes, output_offset


def build_btc_block(
        block_pieces: Deque[Union[bytearray, memoryview]], size: int
) -> Tuple[BlockBtcMessage, int]:
    btc_block = bytearray(size)
    offset = 0
    for piece in block_pieces:
        next_offset = offset + len(piece)
        btc_block[offset:next_offset] = piece
        offset = next_offset
    return BlockBtcMessage(buf=btc_block), offset


def compute_short_id(key: bytes, tx_hash_binary: Union[bytearray, memoryview]) -> bytes:
    return siphash24(key, bytes(tx_hash_binary))[0:6]


class BtcNormalMessageConverter(AbstractBtcMessageConverter):

    def block_to_bx_block(
        self, block_msg, tx_service, enable_block_compression: bool, min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        """
        Compresses a Bitcoin block's transactions and packs it into a bloXroute block.
        """
        compress_start_datetime = datetime.utcnow()
        compress_start_timestamp = time.time()
        size = 0
        buf = deque()
        short_ids = []
        original_size = len(block_msg.rawbytes())
        ignored_sids = []

        header = block_msg.header()
        size += len(header)
        buf.append(header)

        max_timestamp_for_compression = time.time() - min_tx_age_seconds

        for tx in block_msg.txns():
            tx_hash = btc_common_utils.get_txid(tx)
            transaction_key = tx_service.get_transaction_key(tx_hash)
            short_id = tx_service.get_short_id_by_key(transaction_key)

            short_id_assign_time = 0
            if short_id != constants.NULL_TX_SID:
                short_id_assign_time = tx_service.get_short_id_assign_time(short_id)

            if short_id == constants.NULL_TX_SID or \
                    not enable_block_compression or \
                    short_id_assign_time > max_timestamp_for_compression:
                if short_id != constants.NULL_TX_SIDS:
                    ignored_sids.append(short_id)
                buf.append(tx)
                size += len(tx)
            else:
                short_ids.append(short_id)
                buf.append(btc_constants.BTC_SHORT_ID_INDICATOR_AS_BYTEARRAY)
                size += 1

        block = finalize_block_bytes(buf, size, short_ids)

        prev_block_hash = convert.bytes_to_hex(block_msg.prev_block_hash().binary)
        bx_block_hash = convert.bytes_to_hex(crypto.double_sha256(block))

        block_info = BlockInfo(
            block_msg.block_hash(),
            short_ids,
            compress_start_datetime,
            datetime.utcnow(),
            (time.time() - compress_start_timestamp) * 1000,
            block_msg.txn_count(),
            bx_block_hash,
            prev_block_hash,
            original_size,
            size,
            100 - float(size) / original_size * 100,
            ignored_sids
        )

        return memoryview(block), block_info

    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        """
        Uncompresses a bx_block from a broadcast bx_block message and converts to a raw BTC bx_block.

        bx_block must be a memoryview, since memoryview[offset] returns a bytearray, while bytearray[offset] returns
        a byte.
        """
        if not isinstance(bx_block_msg, memoryview):
            bx_block_msg = memoryview(bx_block_msg)

        decompress_start_datetime = datetime.utcnow()
        decompress_start_timestamp = time.time()

        # Initialize tracking of transaction and SID mapping
        block_pieces = deque()
        header_info = parse_bx_block_header(bx_block_msg, block_pieces)
        unknown_tx_sids, unknown_tx_hashes, offset = parse_bx_block_transactions(
            header_info.block_hash,
            bx_block_msg,
            header_info.offset,
            header_info.short_ids,
            header_info.block_offsets,
            tx_service,
            block_pieces
        )
        total_tx_count = header_info.txn_count

        if not unknown_tx_sids and not unknown_tx_hashes:
            btc_block_msg, offset = build_btc_block(block_pieces, offset)
            logger.debug(
                "Successfully parsed block broadcast message. {} transactions "
                "in block {}",
                total_tx_count,
                header_info.block_hash
            )
        else:
            btc_block_msg = None
            logger.debug(
                "Block recovery needed for {}. Missing {} sids, {} tx hashes. "
                "Total txs in block: {}",
                header_info.block_hash,
                len(unknown_tx_sids),
                len(unknown_tx_hashes),
                total_tx_count
            )
        block_info = get_block_info(
            bx_block_msg,
            header_info.block_hash,
            header_info.short_ids,
            decompress_start_datetime,
            decompress_start_timestamp,
            total_tx_count,
            btc_block_msg
        )
        return BlockDecompressionResult(btc_block_msg, block_info, unknown_tx_sids, unknown_tx_hashes)

    def compact_block_to_bx_block(
        self,
        compact_block: CompactBlockBtcMessage,
        transaction_service: TransactionService
    ) -> CompactBlockCompressionResult:
        """
         Handle decompression of Bitcoin compact block.
         Decompression converts compact block message to full block message.
         """
        compress_start_datetime = datetime.utcnow()
        block_header = compact_block.block_header()
        sha256_hash = hashlib.sha256()
        sha256_hash.update(block_header)
        sha256_hash.update(compact_block.short_nonce_buf())
        hex_digest = sha256_hash.digest()
        key = hex_digest[0:16]

        short_ids = compact_block.short_ids()

        short_id_to_tx_contents = {}

        for tx_hash in transaction_service.iter_transaction_hashes():
            transaction_key = transaction_service.get_transaction_key(tx_hash)
            tx_hash_binary = tx_hash.binary[::-1]
            tx_short_id = compute_short_id(key, tx_hash_binary)
            if tx_short_id in short_ids:
                tx_content = transaction_service.get_transaction_by_key(transaction_key)
                if tx_content is None:
                    logger.debug("Hash {} is known by transactions service but content is missing.", tx_hash)
                else:
                    short_id_to_tx_contents[tx_short_id] = tx_content
            if len(short_id_to_tx_contents) == len(short_ids):
                break

        block_transactions = []
        missing_transactions_indices = []
        pre_filled_transactions = compact_block.pre_filled_transactions()
        total_txs_count = len(pre_filled_transactions) + len(short_ids)

        size = 0
        block_msg_parts = deque()

        block_msg_parts.append(block_header)
        size += len(block_header)

        tx_count_size = btc_messages_util.get_sizeof_btc_varint(total_txs_count)
        tx_count_buf = bytearray(tx_count_size)
        btc_messages_util.pack_int_to_btc_varint(total_txs_count, tx_count_buf, 0)
        block_msg_parts.append(tx_count_buf)
        size += tx_count_size

        short_ids_iter = iter(short_ids.keys())

        for index in range(total_txs_count):
            if index not in pre_filled_transactions:
                short_id = next(short_ids_iter)

                if short_id in short_id_to_tx_contents:
                    short_tx = short_id_to_tx_contents[short_id]
                    block_msg_parts.append(short_tx)
                    block_transactions.append(short_tx)
                    size += len(short_tx)
                else:
                    missing_transactions_indices.append(index)
                    block_transactions.append(None)
            else:
                pre_filled_transaction = pre_filled_transactions[index]
                block_msg_parts.append(pre_filled_transaction)
                block_transactions.append(pre_filled_transaction)
                size += len(pre_filled_transaction)

        recovered_item = CompactBlockRecoveryData(
            block_transactions, block_header, compact_block.magic(), transaction_service
        )

        block_info = BlockInfo(
            compact_block.block_hash(),
            [],
            compress_start_datetime,
            compress_start_datetime,
            0,
            None,
            None,
            None,
            len(compact_block.rawbytes()),
            None,
            None,
            []
        )

        if len(missing_transactions_indices) > 0:
            recovery_index = self._last_recovery_idx
            self._last_recovery_idx += 1
            self._recovery_items[recovery_index] = recovered_item  # pyre-ignore
            return CompactBlockCompressionResult(
                False,
                block_info,
                None,
                recovery_index,
                missing_transactions_indices,
                []
            )
        result = CompactBlockCompressionResult(False, block_info, None, None, [], [])
        return self._recovered_compact_block_to_bx_block(result, recovered_item)

    def recovered_compact_block_to_bx_block(
        self,
        failed_compression_result: CompactBlockCompressionResult,
    ) -> CompactBlockCompressionResult:
        failed_block_info = failed_compression_result.block_info
        start_datetime = datetime.utcnow()
        block_info = BlockInfo(
            failed_block_info.block_hash,  # pyre-ignore
            failed_block_info.short_ids,  # pyre-ignore
            start_datetime,
            start_datetime,
            0,
            None,
            None,
            None,
            failed_block_info.original_size,  # pyre-ignore
            None,
            None,
            []
        )
        failed_compression_result.block_info = block_info
        return self._recovered_compact_block_to_bx_block(
            failed_compression_result,
            self._recovery_items.pop(failed_compression_result.recovery_index)  # pyre-ignore
        )

    def _recovered_compact_block_to_bx_block(
        self,
        compression_result: CompactBlockCompressionResult,
        recovery_item: CompactBlockRecoveryData
    ) -> CompactBlockCompressionResult:
        """
        Handle recovery of Bitcoin compact block message.
        """

        missing_indices = compression_result.missing_indices
        recovered_transactions = compression_result.recovered_transactions
        block_transactions = recovery_item.block_transactions
        if len(missing_indices) != len(recovered_transactions):
            logger.debug(
                "Number of transactions missing in compact block does not match number of recovered transactions."
                "Missing transactions - {}. Recovered transactions - {}", len(missing_indices),
                len(recovered_transactions))
            return CompactBlockCompressionResult(
                False, None, None, None, missing_indices, recovered_transactions
            )

        for i in range(len(missing_indices)):
            missing_index = missing_indices[i]
            block_transactions[missing_index] = recovered_transactions[i]

        size = 0
        total_txs_count = len(block_transactions)
        block_msg_parts = deque()

        block_header = recovery_item.block_header
        block_msg_parts.append(block_header)
        size += len(block_header)

        tx_count_size = btc_messages_util.get_sizeof_btc_varint(total_txs_count)
        tx_count_buf = bytearray(tx_count_size)
        btc_messages_util.pack_int_to_btc_varint(total_txs_count, tx_count_buf, 0)
        block_msg_parts.append(tx_count_buf)
        size += tx_count_size

        for transaction in block_transactions:
            block_msg_parts.append(transaction)
            size += len(transaction)  # pyre-ignore

        msg_header = bytearray(btc_constants.BTC_HDR_COMMON_OFF)
        struct.pack_into("<L12sL", msg_header, 0, recovery_item.magic, BtcMessageType.BLOCK, size)
        block_msg_parts.appendleft(msg_header)
        size += btc_constants.BTC_HDR_COMMON_OFF

        block_msg_bytes = bytearray(size)
        off = 0
        for blob in block_msg_parts:
            next_off = off + len(blob)
            block_msg_bytes[off:next_off] = blob
            off = next_off

        checksum = crypto.bitcoin_hash(block_msg_bytes[btc_constants.BTC_HDR_COMMON_OFF:size])
        block_msg_bytes[btc_constants.BTC_HEADER_MINUS_CHECKSUM:btc_constants.BTC_HDR_COMMON_OFF] = checksum[0:4]
        bx_block, compression_block_info = self.block_to_bx_block(
            BlockBtcMessage(buf=block_msg_bytes), recovery_item.tx_service, True, 0
        ) # TODO need to think about a better algorithm
        compress_start_datetime = compression_block_info.start_datetime
        compress_end_datetime = datetime.utcnow()
        block_info = BlockInfo(
            compression_block_info.block_hash,
            compression_block_info.short_ids,
            compress_start_datetime,
            compress_end_datetime,
            (compress_end_datetime - compress_start_datetime).total_seconds() * 1000,
            compression_block_info.txn_count,
            compression_block_info.compressed_block_hash,
            compression_block_info.prev_block_hash,
            compression_block_info.original_size,
            compression_block_info.compressed_size,
            compression_block_info.compression_rate,
            compression_block_info.ignored_short_ids
        )
        return CompactBlockCompressionResult(True, block_info, bx_block, None, [], [])




import re
import socket
import sys
import time
from typing import List, Optional

import urllib3

from bxcommon import constants
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxutils import log_messages
from bxutils import logging

logger = logging.get_logger(__name__)


def blocking_resolve_ip(net_address: str) -> str:
    tries = 0
    resolved_ip = None
    while resolved_ip is None:
        try:
            resolved_ip = socket.gethostbyname(net_address)
        except socket.error:
            time.sleep(constants.NET_ADDR_INIT_CONNECT_RETRY_INTERVAL_SECONDS)
            resolved_ip = None
            tries += 1

            logger.debug("Unable to resolve address {0}. Retried {1}", net_address, tries)
            if tries >= constants.NET_ADDR_INIT_CONNECT_TRIES:
                raise EnvironmentError("Unable to resolve address {}.".format(net_address))
    assert resolved_ip is not None
    return resolved_ip


def blocking_resolve_peers(peer_models: List[OutboundPeerModel]):
    for peer in peer_models:
        resolved_ip = blocking_resolve_ip(peer.ip)

        if peer.ip != resolved_ip:
            logger.trace("Resolved peer {0} to {1}".format(peer.ip, resolved_ip))
            peer.ip = resolved_ip

    logger.trace("Resolved peers successfully.")


def get_node_public_ip() -> Optional[str]:
    """
    Get the public IP address of the node based on a request to `PUBLIC_IP_ADDR_RESOLVER`. If address cannot be
    resolved, it must be specified as a command line argument

    :return: the resolved IP address
    """
    try:
        http = urllib3.PoolManager()
        for _ in range(3):
            result = http.request("GET", constants.PUBLIC_IP_ADDR_RESOLVER)
            if result is not None:
                decoded = result.data.decode("utf-8")
                public_ip_addr = re.findall(constants.PUBLIC_IP_ADDR_REGEX, decoded)
                if public_ip_addr:
                    return public_ip_addr[0]
                else:
                    logger.debug("Unable to parse IP from response - response was {0}", decoded)

        raise ConnectionError("Unable to parse IP from response - no response was returned")

    # pylint: disable=broad-except
    except Exception as e:
        logger.error(log_messages.UNABLE_TO_DETERMINE_PUBLIC_IP, e)
        sys.exit(1)
